import numpy as np
from decimal import *
import json
from copy import deepcopy

class Param(object):
    """
    A parameter in a bayesim analysis. Can be a fitting parameter or an experimental condition.
    """

    def __init__(self, **argv):
        """
        Initialize a Param object.

        Args:
            name (`str`): name of the parameter, required
            units (`str`): units in which parameter is measured (defaults to 'unitless')
            tolerance (`float`): smallest difference between two values of this parameter to consider "real"
            display_name (str): name for plotting (can include TeX), defaults to name
        """
        self.name = argv['name']
        units = argv.get('units','unitless')
        self.units = units
        if 'tolerance' in argv.keys():
            self.set_tolerance(argv['tolerance'])
        else:
            self.tolerance = None
        if 'display_name' in argv.keys():
            self.display_name = argv['display_name']
        else:
            self.display_name = self.name

    def set_tolerance(self, tol, islog=False):
        """Set the tolerance for this parameter."""
        if hasattr(self, 'spacing'):
            if self.spacing=='log':
                self.tolerance = tol
            elif self.spacing=='linear':
                self.tol_digits = int(-1*(np.floor(np.log10(tol))))
                self.tolerance = round(tol, self.tol_digits)
        else:
            self.tol_digits = int(-1*(np.floor(np.log10(tol))))
            self.tolerance = round(tol, self.tol_digits)

    def get_val_str(self, val):
        tol_digits = int(-1*(np.floor(np.log10(self.tolerance))))
        return str(round(val, tol_digits))

class Fit_param(Param):
    """
    A bayesim fitting parameter. Because they will be initialized on a grid, each fitting parameter stores its full list of values as well as some other information such as the spacing between them and the minimum width of a box (used during grid subdivisions).
    """

    def __init__(self, **argv):
        """
        Initialize the fitting parameter.

        Args:
            name (`str`): name of the parameter, required
            units (`str`): units in which parameter is measured (defaults to 'unitless')
            tolerance (`float`): smallest difference between two values of this parameter to consider "real", defaults to 0.1 of min_width
            val_range (:obj:`list` of :obj:`float`): [min, max] (either this or vals is required)
            vals (:obj:`list` of :obj:`float`): full list of vals for this param
            length (int): initial length of this parameter (defaults to 10)
            min_width(`float`): minimum box width for this parameter - subtractive if linear spacing and divisive if logarithmic (defaults to 0.01 of total range, required if providing val_range)
            spacing (str): 'linear' or 'log' (defaults to linear)
            verbose (bool): flag for verbosity
        """
        # get spacing and length (or set defaults)
        if not 'vals' in argv.keys():
            self.spacing = argv.get('spacing','linear')
            self.length = argv.get('length',10)
        Param.__init__(self, **argv) # set name, units, tolerance
        verbose = argv.get('verbose',False)

        # sanity check
        assert ('val_range' in argv or 'vals' in argv), "You must provide a range of values for this fitting parameter!"

        # compute edges and values
        if 'vals' in argv.keys():
            vals = argv['vals']
            assert(len(vals)>=2), "vals should have at least two values"
            vals = sorted(vals)
            self.length = len(vals)
            edges = np.zeros(len(vals)+1)

            if 'spacing' in argv.keys():
                self.spacing = argv['spacing']
            else:
                # try to guess spacing
                diffs = [vals[i+1]-vals[i] for i in range(len(vals)-1)]
                ratios = [vals[i+1]/vals[i] for i in range(len(vals)-1)]
                if np.std(diffs)/np.mean(diffs) < np.std(ratios)/np.mean(ratios):
                    self.spacing = 'linear'
                else:
                    self.spacing = 'log'

            # first edge
            if self.spacing=='linear':
                edges[0] = vals[0]-0.5*(vals[1]-vals[0])
            elif self.spacing=='log':
                edges[0] = vals[0]/((vals[1]/vals[0])**0.5)
            # most of the edges
            for i in range(1,len(vals)):
                if self.spacing=='linear':
                    edges[i] = vals[i]-0.5*(vals[i]-vals[i-1])
                elif self.spacing=='log':
                    edges[i] = vals[i]/((vals[i]/vals[i-1])**0.5)
            # last edge
            if self.spacing=='linear':
                edges[len(vals)] = vals[-1]+0.5*(vals[-1]-vals[-2])
            elif self.spacing=='log':
                edges[len(vals)] = vals[-1]*((vals[-1]/vals[-2])**0.5)

            self.vals = vals
            self.val_range = [min(edges),max(edges)]
            self.edges = edges

        elif 'val_range' in argv.keys():
            self.val_range = argv['val_range']
            assert len(self.val_range)==2 and self.val_range[1]>self.val_range[0], "val_range must be of length 2 with second entry larger than first!"
            if self.spacing == 'linear':
                edges_vals = np.linspace(self.val_range[0],self.val_range[1],2*self.length+1)
            elif self.spacing == 'log':
                edges_vals = np.logspace(np.log10(self.val_range[0]),np.log10(self.val_range[1]),2*self.length+1)
            self.edges = edges_vals[::2]
            self.vals = edges_vals[1:-1:2]

        # set min_width
        if 'min_width' not in argv.keys():
            if verbose:
                print('Setting min_width automatically for %s.'%self.name)
            if self.spacing == 'linear':
                min_width = (1./(10*self.length))*(self.val_range[1]-self.val_range[0])
            elif self.spacing == 'log':
                min_width = (self.val_range[1]/self.val_range[0])**(1./(10*self.length))
        else:
            min_width = argv['min_width']
        self.min_width = min_width

        # set tolerance if not provided
        if 'tolerance' not in argv.keys():
            if self.spacing=='linear':
                self.set_tolerance(0.1*self.min_width)
            elif self.spacing=='log':
                self.set_tolerance(self.min_width**0.1, True)
        else:
            if self.spacing=='linear':
                self.set_tolerance(argv['tolerance'])
            elif self.spacing=='log':
                self.set_tolerance(argv['tolerance'], True)

        # now round all the values based on tolerance
        if self.spacing=='linear': #tolerance is subtractive
            num_digits = self.get_tol_digits()
            self.edges = [round(e, num_digits) for e in self.edges]
            self.vals = [round(v, num_digits) for v in self.vals]
            self.val_range = [round(v, num_digits) for v in self.val_range]
        elif self.spacing=='log': #tolerance is multiplicative
            for i in range(len(self.vals)):
                self.vals[i] = round(self.vals[i], self.get_tol_digits(val=self.vals[i]))
            for i in range(len(self.edges)):
                self.edges[i] = round(self.edges[i], self.get_tol_digits(val=self.edges[i]))
            for i in range(2):
                self.val_range[i] = round(self.val_range[i], self.get_tol_digits(val=self.val_range[i]))

    def get_tol_digits(self, **argv):
        """Compute number of digits to round to. 'val' must be provided if logspaced."""
        if self.spacing=='linear':
            return self.tol_digits
        elif self.spacing=='log':
            assert 'val' in argv.keys(), "If parameter is log-spaced, the value is needed to calculate tolerance digits!"
            val = argv['val']
            tol_val = abs(val - val/self.tolerance)
            return int(-1*(np.floor(np.log10(tol_val))))+1

    def get_closest_val(self, val):
        """Return closest value to val in this parameters current set of vals."""
        diffs = [abs(val-v) for v in self.vals]
        #print(diffs)
        #print(diffs, min(diffs))
        val_ind = diffs.index(min(diffs))
        closest_val = self.vals[val_ind]
        digits = self.get_tol_digits(val=closest_val)
        if abs(round(closest_val, digits-1)/round(val, digits-1)-1.0)>0.01:
            print("The values %f and %f were pretty far apart for %s..."%(val, closest_val, self.name))
        return closest_val

    def get_val_str(self, val):
        """Return a string with this parameter's value, reasonably formatted."""
        # determine if it should be in scientific notation
        if val>1e4 or val<1e-4:
            return "{:.2E}".format(Decimal(str(float(val))))
        else:
            tol_digits = self.get_tol_digits(val=val)
            if tol_digits < 1:
                return str(int(val))
            else:
                return "{0:.{1}f}".format(val, tol_digits)

class Measured_param(Param):
    """A bayesim measured parameter. Can be experimental input or output."""
    def __init__(self, **argv):
        """
        Initialize a Param object.

        Args:
            name (`str`): name of the parameter, required
            units (`str`): units in which parameter is measured (defaults to 'unitless')
            tolerance (`float`): smallest difference between two values of this parameter to consider "real," defaults to 1E-6
            param_type (`str`): 'input' or 'output', defaults to input
        """
        Param.__init__(self, **argv) # set name, units, tolerance
        self.param_type = argv.get('param_type','input')

        # set tolerance if not provided
        if 'tolerance' not in argv.keys():
            self.set_tolerance(1e-6)

class Param_list(object):
    """
    Small class to facilitate listing and comparison of bayesim parameters.
    """

    def __init__(self, **argv):
        """
        Initialize an empty Param_list, or initialize from a dict if provided.

        Args:
            param_dict (`dict`): output of a call to as_dict()
        """
        self.fit_params = []
        self.ecs = []
        self.output = []
        self.ec_x_name = None

        if 'param_dict' in argv.keys():
            param_dict = argv['param_dict']
            for fp in param_dict['fit_params']:
                self.fit_params.append(Fit_param(**fp))
            for ec in param_dict['ecs']:
                self.ecs.append(Measured_param(**ec))
            for o_var in param_dict['output']:
                self.output.append(Measured_param(**o_var))
            self.ec_x_name = param_dict['ec_x_name']

    def add_fit_param(self, **argv):
        """
        Add a fitting parameter to the list.

        Args:
            param (:class:`Fit_param`): A Fit_param object to add to the list
            name (str): name of the parameter, required if param object not passed
            units (str): units in which parameter is measured (defaults to 'unitless')
            tolerance (float): smallest difference between two values of this parameter to consider "real"
            val_range (`:obj:`list`` of :obj:`float`): [min, max] (either this or vals is required)
            vals (:obj:`list` of :obj:`float`): full list of vals for this param
            length (int): initial length of this parameter (defaults to 10)
            min_width(float): minimum box width for this parameter - subtractive if linear spacing and divisive if logarithmic (defaults to 0.01 of total range, required if providing val_range)
            spacing (str): 'linear' or 'log' (defaults to linear)
            verbose (bool): verbosity flag
        """
        if 'name' in argv.keys():
            name = argv['name']
        elif 'param' in argv.keys():
            name = argv['param'].name
        else:
            raise ValueError("You must provide either a parameter name or a Fit_param object!")
        verbose = argv.get('verbose',False)

        if not self.param_present(name):
            self.fit_params.append(Fit_param(**argv))
        else: # overwrite
            if verbose:
                print("Overwriting metadata for fitting parameter %s with new info."%name)
            param_ind = self.fit_params.index(self.find_param(name))
            if 'param' in argv.keys():
                self.fit_params[param_ind] = argv['param']
            else:
                #print(argv)
                self.fit_params[param_ind] = Fit_param(**argv)

    def add_ec(self, **argv):
        """
        Add an experimental condition.

        Args:
            name (str): name of the parameter, required
            units (str): units in which parameter is measured (defaults to 'unitless')
            tolerance (float): smallest difference between two values of this parameter to consider "real," defaults to 1E-6
            is_x (bool): set this to be the x-axis variable when plotting data, defaults to False
        """
        if not self.param_present(argv['name']):
            args = dict(argv)
            args['param_type'] = 'input'
            self.ecs.append(Measured_param(**args))
            is_x = argv.get('is_x', False)
            if is_x:
                self.set_ec_x(argv['name'])
        else:
            raise ValueError("It looks like you're trying to add a duplicate experimental condition, %s!"%argv['name'])

    def set_ec_x(self, param_name, verbose=False):
        """Set the x-variable for experimental conditions."""
        if not param_name in self.param_names('ec'):
            if verbose:
                print("Adding the variable %s to the list of experimental conditions and setting it as the x-axis variable for plotting."%param_name)
            self.add_ec(name=param_name)
        elif self.ec_x_name==param_name:
            pass #already done
        elif not self.ec_x_name==None:
            print("Overwriting previous ec_x variable (%s) with %s."%(self.ec_x_name,param_name))
        self.ec_x_name = param_name

    def get_ec_x(self):
        return self.find_param(self.ec_x_name)

    def add_output(self, **argv):
        """
        Add an output variable.

        Args:
            name (str): name of the parameter, required
            units (str): units in which parameter is measured (defaults to 'unitless')
            tolerance (float): smallest difference between two values of this parameter to consider "real," defaults to 1E-6
        """
        if not self.param_present(argv['name']):
            args = dict(argv)
            args['param_type'] = 'output'
            self.output.append(Measured_param(**args))
        else:
            raise ValueError("It looks like you're trying to add a duplicate output variable, %s!"%argv['name'])

    def param_names(self, param_type=None):
        """
        Return a list of parameter names. If no arguments provided, output will be a dict, if a type is provided, output will be a list of just the parameter names of that type.
        """
        if param_type=='fit':
            return [p.name for p in self.fit_params]
        elif param_type=='ec':
            return [c.name for c in self.ecs]
        elif param_type=='output':
            return [o.name for o in self.output]
        else:
            return {'fit_params': [p.name for p in self.fit_params],
                    'ec': [p.name for p in self.ec],
                    'output': [p.name for p in self.output]}

    def all_params(self):
        """Return a flat list of all parameters of any type."""
        return [p for p in self.fit_params+self.ecs+self.output]

    def param_present(self, name):
        """
        Check that the param name isn't already present in a list.

        Args:
            name (str): name to check for
        """
        all_names = [p.name for p in self.all_params()]
        if name in all_names:
            return True
        else:
            return False

    def find_param(self, name):
        """
        Return the Param object with the given name.

        Args:
            name (str): name to search for
        """
        return [p for p in self.all_params() if p.name==name][0]

    def vals_equal(self, param_name, val1, val2):
        """
        Compare two values of a given param.

        Args:
            param_name (str): name of parameter, must be in one of the lists
            val1, val2 (float): values to be compared

        Returns:
            True if abs(val1-val2) < tolerance of param_name
        """
        tol = self.find_param(param_name)['tolerance']
        if abs(val1-val2)<=tol:
            return True
        else:
            return False

    def set_tolerance(self, param_name, tol):
        """Set the tolerance value for the given parameter."""
        self.find_param(param_name).set_tolerance(tol)

    def is_empty(self):
        if len(self.all_params())==0:
            return True
        else:
            return False

    def as_dict(self):
        d = {'ec_x_name':self.ec_x_name}
        d['ecs'] = [c.__dict__ for c in self.ecs]
        d['fit_params'] = [p.__dict__ for p in self.fit_params]
        d['output'] = [o.__dict__ for o in self.output]
        return d

    def __str__(self):
        d = deepcopy(self.as_dict())
        for i in range(len(d['fit_params'])):
            d['fit_params'][i]['edges'] = str(d['fit_params'][i]['edges'])[:70]+'...'
            d['fit_params'][i]['vals'] = str(d['fit_params'][i]['vals'])[:70]+'...'
            d['fit_params'][i]['val_range'] = str(d['fit_params'][i]['val_range'])
        return json.dumps(d, indent=4)



from copy import deepcopy
import numpy as np
import pandas as pd
import sys
sys.path.append('../../')
import matplotlib.pyplot as plt
import matplotlib.patches as mplp
from matplotlib import colors as mcolors
from bisect import bisect_left

def visualize_PMF_sequence(statefile_list, **argv):
    """
    Create plot akin to that produced by pmf.visualize() but with data from multiple PMF's overlaid. All should have the same set of fitting parameters. For now assumes that first statefile has the largest axes bounds, will add automated check later.

    Args:
        statefile_list (list of str): list of paths to statefiles (saved by Model.save_state function storing PMF's to be visualized), in order from least to most subdivided
        name_list (list of str): optional, list of names for legend, defaults to filenames
        true_vals (`dict`): optional, set of param values to highlight on PMF
        fpath (str): optional, path to save image to
    """
    import model as bym

    # get legend names
    if 'name_list' in argv.keys():
        name_list = argv['name_list']
    else:
        name_list = [sf.split('/')[-1].split('.')[0] for sf in statefile_list]

    if 'true_vals' in argv.keys():
        plot_true_vals = True
        true_vals = argv['true_vals']
    else:
        plot_true_vals = False

    # load the models one by one (to save RAM) and get visualizations
    figs = []
    axes_sets = []
    data_sets = []
    for i in range(len(statefile_list)):
        statefile = statefile_list[i]
        m = bym.Model(load_state=True, state_file=statefile)
        if plot_true_vals:
            d = m.probs.visualize(return_plots=True, true_vals=true_vals, color_index=i)
        else:
            d = m.probs.visualize(return_plots=True, color_index=i)
        figs.append(d['fig'])
        axes_sets.append(d['axes'])
        data_sets.append(d['data'])
    num_params = len(m.params.fit_params)

    # get color cycle and colormaps (just pull five)
    prop_cycle = plt.rcParams['axes.prop_cycle']
    colors = prop_cycle.by_key()['color'][:5]
    #cmap_name_list = ['Blues','Greens','Reds','Purples','Oranges']

    # initialize the figure
    fig, axes = plt.subplots(nrows=num_params, ncols=num_params, figsize=(5*num_params, 5*num_params))

    for rownum in range(num_params):
        for colnum in range(num_params):
            x_param = m.params.fit_params[colnum]
            y_param = m.params.fit_params[rownum]
            if plot_true_vals:
                true_x = true_vals[x_param.name]
                true_y = true_vals[y_param.name]
                if x_param.spacing=='log':
                    true_x = np.log10(true_x)
                if y_param.spacing=='log':
                    true_y = np.log10(true_y)
            for i in range(len(figs)):
                old_ax = axes_sets[i][rownum][colnum]
                #color = colors[i]
                #patches = old_ax.patches
                if i==0: # formatting stuff
                    if rownum<colnum: # build the legend here
                        if rownum==0 and colnum==num_params-1:
                            axes[rownum][colnum].axis('off')
                            axes[rownum][colnum].set_xlim([0,1])
                            axes[rownum][colnum].set_ylim([0,1])

                            for j in range(len(axes_sets)):
                                x = 0.2
                                if plot_true_vals:
                                    y = 0.8-(float(j)/float(len(axes_sets)+1))*0.7
                                else:
                                    y = 0.8-(float(j)/float(len(axes_sets)))*0.7
                                axes[rownum][colnum].add_patch(mplp.Rectangle((x,y), 0.1, 0.1, facecolor=colors[j]))
                                axes[rownum][colnum].text(x+0.15, y+0.01, name_list[j], color='k', fontsize=20)
                            if plot_true_vals:
                                axes[rownum][colnum].add_patch(mplp.Rectangle((0.13,0.06), 0.17, 0.08, facecolor='0.8'))
                                axes[rownum][colnum].scatter([0.16],[0.1],200, '#FFFF00', marker='*', zorder=200)
                                axes[rownum][colnum].scatter([0.25],[0.1],200, c="None", marker='o', linewidths=3, edgecolors='#FFFF00', zorder=200)
                                axes[rownum][colnum].text(0.35, 0.08, 'ground truth', color='k', fontsize=20)
                        else:
                            fig.delaxes(axes[rownum][colnum])
                    else: # set up the axes for plotting
                        axes[rownum][colnum].set_xlim(old_ax.get_xlim())
                        axes[rownum][colnum].set_ylim(old_ax.get_ylim())
                        axes[rownum][colnum].set_xlabel(old_ax.get_xlabel(), fontsize=24)
                        axes[rownum][colnum].set_xscale(old_ax.get_xscale())
                        axes[rownum][colnum].set_yscale(old_ax.get_yscale())
                        axes[rownum][colnum].set_xticks(old_ax.get_xticks())
                        axes[rownum][colnum].set_yticks(old_ax.get_yticks())
                        axes[rownum][colnum].set_xticklabels(old_ax.get_xticklabels())
                        axes[rownum][colnum].set_yticklabels(old_ax.get_yticklabels())
                        axes[rownum][colnum].set_axisbelow(True)

                        if rownum==colnum:
                            axes[rownum][colnum].yaxis.set_label_position("right")
                            axes[rownum][colnum].set_ylabel(old_ax.get_ylabel(), rotation=270, labelpad=24, fontsize=24)
                            if plot_true_vals:
                                if x_param.spacing=='log':
                                    axes[rownum][colnum].scatter(10.0**true_x,0.05,200,'#FFFF00',marker='*',zorder=100)
                                elif x_param.spacing=='linear':
                                    axes[rownum][colnum].scatter(true_x,0.05,200,'#FFFF00',marker='*',zorder=100)

                        elif rownum > colnum:
                            axes[rownum][colnum].set_ylabel(old_ax.get_ylabel(), fontsize=24)
                            if plot_true_vals:
                                #true_x, true_y = old_ax.collections[0]._offsets[0]
                                axes[rownum][colnum].scatter(true_x,true_y,200,c="None",marker='o',linewidths=3,edgecolors='#FFFF00',zorder=100)

                        for item in (axes[rownum][colnum].get_xticklabels() + axes[rownum][colnum].get_yticklabels()):
                            item.set_fontsize(20)

                if rownum==colnum:
                    hist_data = data_sets[i][rownum][colnum]
                    axes[rownum][colnum].hist(hist_data['vals'], weights=hist_data['weights'], bins=hist_data['bins'], color=hist_data['color'],zorder=50+i)
                elif rownum > colnum:
                    img_data = data_sets[i][rownum][colnum]
                    axes[rownum][colnum].imshow(img_data['image'], aspect='auto', origin='lower', extent=img_data['extent'], zorder=50+i)

    if 'fpath' in argv.keys():
        plt.savefig(argv['fpath'])

    else:
        plt.show()

def get_closest_val(val, val_list):
    """Return closest value to val in a list."""
    pos = bisect_left(val_list, val)
    if pos == 0:
        return val_list[0]
    if pos == len(val_list):
        return val_list[-1]
    before = val_list[pos - 1]
    after = val_list[pos]
    if after - val < val - before:
       return after
    else:
       return before

def calc_deltas(grp, inds, param_lengths, model_data, fit_param_names, probs, output_var, take_average):
    # construct matrix of output_var({fit_params})
    subset = deepcopy(model_data.loc[inds])
    # sort and reset index of subset to match probs so we can use the find_neighbor_boxes function if needed
    subset.drop_duplicates(subset=fit_param_names, inplace=True)
    subset.sort_values(fit_param_names, inplace=True)
    subset.reset_index(inplace=True)
    # removing this check for now and hopefully the missing data filters will handle it
    #if not len(subset.index)==len(probs.points.index):
        #raise ValueError("Subset at EC's %s does not match probability grid!"%str(grp))

    # check if on a grid
    if not len(subset)==np.product(param_lengths):
        is_grid = False
        # construct grid at the highest level of subdivision
        dense_grid = probs.populate_dense_grid(df=subset, col_to_pull=output_var, make_ind_lists=True)
        mat = dense_grid['mat']
        ind_lists = dense_grid['ind_lists']

    else:
        is_grid = True
        mat = np.reshape(list(subset[output_var]), param_lengths)

    # given matrix, compute largest differences along any direction
    winner_dim = [len(mat.shape)]
    winner_dim.extend(mat.shape)
    winners = np.zeros(winner_dim)

    # for every dimension (fitting parameter)
    for i in range(len(mat.shape)):
        # build delta matrix
        # certain versions of numpy throw an "invalid value encountered" RuntimeError here but the function behaves correctly
        with np.errstate(invalid='ignore'):
            deltas_here = np.absolute(np.diff(mat,axis=i))
        pad_widths = [(0,0) for j in range(len(mat.shape))]
        pad_widths[i] = (1,1)
        deltas_here = np.pad(deltas_here, pad_widths, mode='constant', constant_values=0)

        # build "winner" matrix in this direction (ignore nans)
        # this is really ugly because we have to index in at variable positions...
        # likewise here with the error
        with np.errstate(invalid='ignore'):
            winners[i]=np.fmax(deltas_here[tuple([Ellipsis]+[slice(None,mat.shape[i],None)]+[slice(None)]*(len(mat.shape)-i-1))],deltas_here[tuple([Ellipsis]+[slice(1,mat.shape[i]+1,None)]+[slice(None)]*(len(mat.shape)-i-1))])

    grad = np.amax(winners, axis=0)
    if take_average:
        avg = np.nanmean(grad) # compute average
        grad = avg * np.ones(grad.shape) # replace every entry with average

    # save these values to the appropriate indices in the vector
    if is_grid:
        to_return = (grp, grad.flatten())
    else:
        # pick out only the boxes that exist
        to_return = (grp, grad[tuple([i for i in list([ind_lists[p] for p in fit_param_names])])])

    # check for NaNs
    to_check = to_return[1]
    if np.any(np.isnan(to_check)):
        if np.all(np.isnan(to_check)):
            # uh-oh
            raise ValueError("All the uncertainties here (" + str(grp) + ") are NaN! Help!")
        else:
            # replace NaN values with average of non-NaN values for this group
            avg_unc = np.nanmean(to_check)
            fixed = np.where(np.isnan(to_check), to_check, avg_unc)
            to_return[1] = fixed

    return to_return



import argparse
import bayesim.param_list as pml
import bayesim.model as bym

def main(args=None):
    parser = argparse.ArgumentParser()

    # required values / analysis actions
    parser.add_argument('-o', '-output_variable', help='output variable')
    parser.add_argument('-obs', '-attach_observation', help='attach observation')
    parser.add_argument('-ec_x', '-ec_x_var', help='x-axis variable for data plot')
    parser.add_argument('-mod', '-attach_model', help='attach model')
    parser.add_argument('-run', help='run analysis', action='store_true')
    parser.add_argument('-sub', '-subdivide', help='subdivide', action='store_true')

    # optional stuff / tweaking behavior
    parser.add_argument('-name', '-state_name', help='state name')
    parser.add_argument('-max_step', '-max_ec_x_step', help='largest step in x-axis EC to allow')
    parser.add_argument('-save_step', default=10, help='Steps between two consecutive saved probabilities ')
    parser.add_argument('-th_pm', default=0.8, help='threshold in the probability mass')
    parser.add_argument('-th_pv', default=0.05, help='threshold in parameter space volume')
    parser.add_argument('-prb', '-plot_probability', help='plot probability', action='store_true')


    #args = {k:v for k,v in vars(parser.parse_args()).items() if not vars(parser.parse_args())[k]==None}
    args = parser.parse_args()
    state_name =  vars(args).setdefault('state_name','bayesim_state.h5')

    # define output variable (eventually allow multiple)
    #if 'o' in vars(args).keys():
    if not vars(args)['o']==None:
        if not vars(args)['ec_x']==None:
            m = bym.model(output_var=args.o, ec_x_var=args.ec_x)
        else:
            m=bym.model(output_var=args.o)
        m.save_state()

    # attach observed data
    #if 'obs' in vars(args).keys():
    if not vars(args)['obs']==None:
        m=bym.model(load_state=True, state_name=state_name)
        if not vars(args)['max_step']==None:
            m.attach_observations(fpath=args.obs, max_ec_x_step=args.max_step)
        else:
            m.attach_observations(fpath=args.obs)
        m.save_state()

    # attach modeled data
    #if 'mod' in vars(args).keys():
    if not vars(args)['mod']==None:
        m=bym.model(load_state=True, state_name=state_name)
        m.attach_model(fpath=args.mod, mode='file')
        m.calc_model_errors()
        m.save_state()

    # run inference
    if args.run:
        m=bym.model(load_state=True, state_name=state_name)
        m.run(**vars(args))
        m.save_state()

    # subdivide the grid
    if args.sub:
        m=bym.model(load_state=True, state_name=state_name)
        m.subdivide(**vars(args))
        m.save_state()

    # plot probabilities
    if args.prb:
        m=bym.model(load_state=True, state_name=state_name)
        m.visualize_probs(save_file=True)


if __name__ == "__main__":

    main()



#import sys
#sys.path.append('../')
from bayesim.pmf import Pmf
import bayesim.params as pm
from bayesim.utils import calc_deltas, get_closest_val
import pandas as pd
from copy import deepcopy
import numpy as np
import matplotlib.pyplot as plt
import random
import os
from joblib import Parallel, delayed, cpu_count
import time
import platform

class Model(object):
    """
    The main workhorse class of bayesim. Stores the modeled and observed data as well as a Pmf object which maintains the current probability distribution and grid subdivisions.

    Attributes:
    [update this]
    """

    def __init__(self,**argv):
        """
        Initialize with a uniform PMF over the fitting parameters.

        Args:
            obs_data_path (`str`): path to HDF5 file containing measured data
            load_state (`bool`): flag for whether to load state from a file - if True, other inputs (apart from state_file) are ignored
            state_file (`str`): path to file saved by save_state() fcn
            verbose (`bool`): flag for verbosity, defaults to False
            output_var (`str`): name of experimental output measurements
            params (:obj:`param_list`): Param_list object containing parameters to be fit and associated metadata
            ec_x_var (`str`): EC to plot on x-axis and consider in trimming data
            ec_list (:obj:`list` of :obj:`str`): names of experimental conditions
            ec_tols (`dict`): dict of form {ec_name_1:tolerance_1, ec_name_2:tolerance_2, ...}, will supersede ec_list
            ec_units (`dict`): dict of form {ec_name_1:units_1, ec_name_2:units_2, ...}, optional
            model_data_path (`str`): path to HDF5 file containing modeled data
            model_data_func (callable): handle to function for computing model data
        """
        verbose = argv.get('verbose', False)
        state_file = argv.get('state_file', 'bayesim_state.h5')
        load_state = argv.get('load_state', False)

        if load_state:
            if verbose:
                print('Loading bayesim state from %s...\n'%state_file)
            state = pd.read_hdf(state_file)

            # variables
            self.ec_pts = state['ec_pts']
            self.output_var = state['output_var']
            self.params = pm.Param_list(param_dict=state['params'])

            # probabilities
            self.probs = Pmf(prob_dict=state['probs'])

            # model
            self.model_data = state['model_data']
            self.model_data_grps = state['model_data_grps']
            self.model_data_ecgrps = state['model_data_ecgrps']
            self.needs_new_model_data = state['needs_new_model_data']
            self.obs_data = state['obs_data']
            self.is_run = state['is_run']

        else:
            if verbose:
                print('Constructing bayesim model object...\n')
            # initialize empty parameter list
            self.params = pm.Param_list()

            # if a param_list object has been provided, attach it
            if 'params' in argv.keys():
                self.attach_params(argv['params'])
            else:
                self.probs = Pmf()

            # check if output list is populated / attach output variable
            output_list = self.params.output
            if len(output_list)>0: # outputs are populated
                if len(output_list)>1 or ('output_var' in argv.keys() and not output_list[0].name==argv['output_var']):
                    raise ValueError('It seems you are trying to add more than one output variable. Sorry, the current version of bayesim supports only one type of output - this will be addressed in future versions!')
                else:
                    self.output_var = output_list[0].name # for now
            elif 'output_var' in argv.keys():
                self.params.add_output(name=argv['output_var'])
                self.output_var = argv['output_var'] # for now...eventually need to be general to multiple potential outputs
            else:
                raise ValueError("You need to define your output variable!")

            # attach EC's if provided explicitly, or infer from obs_data
            if len(set(['ec_list','ec_tols','ec_units']).intersection(set(argv.keys())))>0:
                self.attach_ecs(**argv)

            # set x-axis EC variable
            if 'ec_x_var' in argv.keys():
                self.params.set_ec_x(argv['ec_x_var'])

            # attach observed data if provided
            if 'obs_data_path' in argv.keys():
                self.attach_observations(**argv)
            else:
                self.ec_pts = pd.DataFrame()
                self.obs_data = pd.DataFrame

            # attach modeled data if provided
            if 'model_data_path' in argv.keys():
                self.attach_model(mode='file', **argv)
            elif 'model_data_func' in argv.keys():
                self.attach_model(mode='function', **argv)
            else:
                self.model_data = pd.DataFrame()
                self.model_data_grps = []
                self.model_data_ecgrps = []
                self.needs_new_model_data = True

            # run flag
            self.is_run = False

    def attach_params(self, params):
        """Attach a param_list object."""
        if not self.params.is_empty():
            print('Overwriting preexiting parameter list with this new one.\n')
        self.params = params
        self.probs = Pmf(params=params.fit_params)

    def attach_ecs(self, **argv):
        """
        Define parameters for experimental conditions.

        Args:
            ec_list (:obj:`list` of :obj:`str`): names of experimental conditions
            ec_tols (`dict`): dict of form {ec_name_1:tolerance_1, ec_name_2:tolerance_2, ...}, will supersede ec_list
            ec_units (`dict`): dict of form {ec_name_1:units_1, ec_name_2:units_2, ...}, optional
        """
        ec_names = []
        tol_dict = {}
        unit_dict = {}
        if 'ec_tols' in argv.keys():
            tol_dict = argv['ec_tols']
            ec_names = tol_dict.keys()
        if 'ec_units' in argv.keys():
            unit_dict = argv['ec_units']
            if ec_names==[]:
                ec_names = unit_dict.keys()
        if 'ec_list' in argv.keys():
            if ec_names==[]:
                ec_names = argv['ec_list']

        for name in ec_names:
            args = {'name':name}
            if name in tol_dict.keys():
                args['tolerance'] = tol_dict[name]
            if name in unit_dict.keys():
                args['units'] = unit_dict[name]
            self.params.add_ec(**args)

    def attach_fit_params(self,params):
        """
        Attach list of parameters to fit.

        Args:
            param_list: list of Fit_param objects
        """
        for param in params:
            self.params.add_fit_param(param=param)

    def attach_observations(self, **argv):
        """
        Attach measured dataset.

        Args:
            obs_data_path (`str`): path to HDF5 file containing observed data
            keep_all (`bool`): whether to keep all the data in the file (longer simulation times) or to clip out data points that are close to each other (defaults to False)
            ec_x_var (`str`): required if keep_all is False, the experimental condition over which to measure differences (e.g. V for JV(Ti) curves in PV). It will also be used in plotting later.
            max_ec_x_step (`float`): used if keep_all is False, largest step to take in the ec_x_var before keeping a point even if curve if "flat" (defaults to 0.05 * range of ec_x_var)
            thresh_dif_frac (`float`): used if keep_all is False, threshold (as a percentage of the range of values, defaults to 0.01)
            fixed_unc (`float`): required if running in function mode or if file doesn't have an 'uncertainty' column, value to use as uncertainty in measurement
            output_column (`str`): optional, header of column containing output data (required if different from self.output_var)
            verbose (`bool`): flag for verbosity, defaults to False
        """
        output_col = argv.get('output_column', self.output_var)
        keep_all = argv.get('keep_all', False)
        thresh_dif_frac = argv.get('thresh_dif_frac', 0.01)
        verbose = argv.get('verbose', False)

        if 'ec_x_var' in argv.keys():
            self.params.set_ec_x(argv['ec_x_var'])
        elif not keep_all and self.params.ec_x_name==None:
            raise NameError('You must specify ec_x_var if you want to throw out data points that are too close together.')

        if verbose:
            print('Attaching measured data...\n')

        self.obs_data = pd.read_hdf(argv['obs_data_path'])

        # get EC names if necessary
        cols = list(self.obs_data.columns)
        if output_col not in cols:
            raise NameError('Your output variable name, %s, is not the name of a column in your input data!' %(output_col))
            return
        elif 'fixed_unc' not in argv.keys() and 'uncertainty' not in cols:
            raise NameError('You need to either provide a value for fixed_unc or your measurement data needs to have an uncertainty column!')
            return
        else:
            cols.remove(output_col)
            if len(self.params.ecs)==0 or ('ec_x_var' in argv.keys() and len(self.params.ecs)==1):
                print('Determining experimental conditions from observed data...\n')
                for c in cols:
                    if not c=='uncertainty':
                        # 1% of the smallest difference between values
                        tol=0.01*min(abs(np.diff(list(set(self.obs_data[c])))))
                        if c==argv['ec_x_var']:
                            self.params.set_tolerance(c, tol)
                        else:
                            self.params.add_ec(name=c, tolerance=tol)
                print('Identified experimental conditions as %s. (If this is wrong, rerun and explicitly specify them with attach_ec (make sure they match data file columns) or remove extra columns from data file.)\n' %(str(self.params.param_names('ec'))))

            # these next bits used to be under an else...
            if 'uncertainty' not in cols:
                self.obs_data['uncertainty'] = argv['fixed_unc']*np.ones(len(self.obs_data))
                cols.extend('uncertainty')
                #print(self.obs_data.head())
                self.params.set_tolerance(self.output_var, 0.01*argv['fixed_unc'])
            else: # set tolerance to 1% of minimum uncertainty
                self.params.set_tolerance(self.output_var, 0.01*min(self.obs_data['uncertainty']))

            if set(cols) == set(self.ec_names()+['uncertainty']):
                pass # all good
            elif set(self.ec_names()+['uncertainty']) <= set(cols):
                print('Ignoring extra columns in data file: %s\n'%(str(list(set(cols)-set(self.ec_names())))))
            elif set(cols) <= set(self.ec_names()+['uncertainty']):
                print('These experimental conditions were missing from your data file: %s\nProceeding assuming that %s is the full set of experimental conditions...\n'%(str(list(set(ec)-set(cols))), str(cols)))
                for c in [c for c in cols if not c=='uncertainty']:
                    self.params.add_ec(name=c)
            # ...else ended here

        # pick out rows to keep - the way to do this thresholding should probably be tweaked
        if not keep_all:
            if verbose:
                print('Choosing which measured data to keep...')
            other_ecs = [ec for ec in self.ec_names() if not ec==self.params.ec_x_name]
            obs_data_grps = self.obs_data.groupby(by=other_ecs)
            for grp in obs_data_grps.groups.keys():
                subset = deepcopy(self.obs_data.loc[obs_data_grps.groups[grp]]).sort_values(self.params.ec_x_name)
                if 'max_ec_x_step' in argv.keys():
                    max_step = argv['max_ec_x_step']
                else:
                    max_step = 0.1 * max(subset[self.params.ec_x_name]-min(subset[self.params.ec_x_name]))
                    print('Using %.2f as the maximum step size in %s when choosing observation points to keep at %s=%s.\n'%(max_step, self.params.ec_x_name, other_ecs, grp))
                thresh = thresh_dif_frac * (max(subset[self.output_var])-min(subset[self.output_var]))
                i = 0
                while i < len(subset)-1:
                    this_pt = subset.iloc[i]
                    next_pt = subset.iloc[i+1]
                    if next_pt[self.params.ec_x_name]-this_pt[self.params.ec_x_name] >= max_step:
                        i = i+1
                    elif next_pt[self.output_var]-this_pt[self.output_var] < thresh:
                        subset.drop(next_pt.name,inplace=True)
                        self.obs_data.drop(next_pt.name,inplace=True)
                    else:
                        i = i+1

        # round EC values
        rd_dct = {c.name:c.tol_digits for c in self.params.ecs}
        self.obs_data = self.obs_data.round(rd_dct)

        # sort observed data
        self.obs_data.sort_values(by=self.ec_names(), inplace=True)

        # populate list of EC points
        if len(self.params.ecs)==1: # it's finicky in this case
            self.ec_pts = pd.DataFrame()
            self.ec_pts[self.ec_names()[0]] = self.obs_data.groupby(self.ec_names()).groups.keys()
        else:
            self.ec_pts = pd.DataFrame.from_records(data=[list(k) for k in self.obs_data.groupby(self.ec_names()).groups.keys()], columns=self.ec_names())

        self.ec_pts = self.ec_pts.sort_values(self.ec_names()).reset_index(drop=True)

    def check_data_columns(self,**argv):
        """
        Make sure the columns in imported data make sense.

        Args:
            model_data (`DataFrame`): dataset to check
            output_column (`str`): optional, header of column containing output data (required if different from self.output_var)
        """
        model_data = argv['model_data']
        output_col = argv['output_column']

        cols = list(model_data.columns)
        # first, check that the output is there
        if output_col not in cols:
            raise NameError('Your output variable name, %s, is not the name of a column in your model data!' %(output_col))
            return
        else:
            cols.remove(output_col)
            # next, that all the experimental conditions are there
            for c in self.ec_names():
                if c not in cols:
                    raise NameError('Experimental condition %s is not the name of a column in your model data!' %(c))
                    return
                else:
                    cols.remove(c)
            # if param_names has been populated, check if they match
            if not self.fit_param_names() == []:
                if set(self.fit_param_names())==set(cols):
                    pass # all good
                elif set(self.fit_param_names()) <= set(cols):
                    print('Ignoring extra columns in model data file: %s\n'%(str(list(set(cols)-set(self.fit_param_names())))))
                elif set(cols) <= set(self.fit_param_names()):
                    print('These experimental conditions were missing from your model data file: %s\nProceeding assuming that %s is the full set of experimental conditions...\n'%(str(list(set(self.fit_param_names())-set(cols))), str(cols)))
            else:
                print("Determining fitting parameters from modeled data...\n")
                for c in cols:
                    if not c=='uncertainty':
                        vals = list(set(model_data[c]))
                        self.params.add_fit_param(name=c, vals=vals)
                print("Found fitting parameters: %s"%self.fit_param_names())

    def check_ecs(self,**argv):
        """
        Check that all experimental conditions are present at each parameter point in modeled data.

        Args:
            gb (:obj:`groupby`): Pandas groupby object of model data grouped by parameter points
            verbose (`bool`): flag for verbosity, defaults to False
        """
        verbose = argv.get('verbose', False)
        if verbose:
            print('Checking that modeled data contains all experimental conditions at every combination of fit parameters...\n')

        grps = argv['gb']
        # then check at each model point that they match
        for name,group in grps:
            #print(self.ec_pts,group[self.ec_names()].sort_values(self.ec_names()).reset_index(drop=True))
            # specifically, we check that EC pts is a SUBSET of model EC's at each parameter space point
            ec_inds = self.ec_pts.index
            if not all(self.ec_pts==group[self.ec_names()].sort_values(self.ec_names()).reset_index(drop=True).loc[ec_inds]):
                # FIX MEEEEE
                #raise ValueError('The experimental conditions do not match to %d digits between the observed and modeled data at the modeled parameter space point %s!'%(self.ec_tol_digits,name))
                print('there is a problem I need to fix the error message for!')
                return

    def calc_indices(self):
        """
        Compute starting and ending indices in self.model_data for each point in self.probs.
        """
        start_indices = np.zeros(len(self.probs.points),dtype=int)
        end_indices = np.zeros(len(self.probs.points),dtype=int)
        #print(list(self.model_data_grps.groups.keys())[:5])
        for pt in self.probs.points.iterrows():
            #print(pt[1])
            #print(tuple(pt[1][self.fit_param_names()].tolist()))
            subset_inds = self.model_data_grps.groups[tuple(pt[1][self.fit_param_names()].tolist())]
            if len(subset_inds)==0:
                print('Something went wrong calculating sim indices! Could not find any points in model data for params %s.'%pt)
            start_ind = int(min(subset_inds))
            end_ind = int(max(subset_inds))
            start_indices[pt[0]] = start_ind
            end_indices[pt[0]] = end_ind
        self.probs.points['start_ind'] = start_indices
        self.probs.points['end_ind'] = end_indices

    def attach_model(self,**argv):
        """
        Attach the model for the data, either by feeding in a file of precomputed data or a function that does the computing.

        Args:
            mode (`str`): 'file' or 'function'
            model_data_func (callable): if mode='function', provide function here
            model_data_path (`str`): if mode=='file', provide path to file
            output_column (`str`): optional, header of column containing output data (required if different from self.output_var)
            calc_model_unc (`bool`): whether to calculate model uncertainties as well, defaults to False
            verbose (`bool`): flag for verbosity, defaults to False
            in_parallel (`bool`): defaults to True, set False to force run in serial
        """
        mode = argv['mode']
        output_col = argv.get('output_column',self.output_var)
        verbose = argv.get('verbose', False)
        calc_model_unc = argv.get('calc_model_unc', False)
        in_parallel = argv.get('in_parallel', True)
        if platform.system()=='Windows': #joblib doesn't work on Windows
            in_parallel = False
        
        if verbose:
            print('Attaching simulated data...')

        if mode == 'file':
            # import and sort data on parameter values
            self.model_data = pd.read_hdf(argv['model_data_path'])

            # Check that columns match EC's and parameter names
            # (and determine parameter names if need be)
            self.check_data_columns(model_data=self.model_data, output_column=output_col)

            # now sort data by param names and EC's
            self.model_data = self.model_data.sort_values(self.fit_param_names()+self.ec_names()).reset_index(drop=True)

            # next get list of parameter space points
            param_points_grps = self.model_data.groupby(self.fit_param_names())
            param_points = pd.DataFrame.from_records(data=[list(k) for k in param_points_grps.groups.keys()],columns=self.fit_param_names()).sort_values(self.fit_param_names()).reset_index(drop=True)

            # if PMF has been populated, check that points match
            # this check is broken ("Can only compare identically-labeled...")
            # FIX ME
            """
            if not self.probs.is_empty:
                #print(len(param_points),len(self.probs.points[self.fit_param_names()]))
                if not all(param_points==self.probs.points[self.fit_param_names()]):
                    print('Your previously populated PMF does not have the same set of parameter space points as your model data. Proceeding using the points from the model data.')
                    self.probs = Pmf()
            """
            ## check that all EC's are present at all model points
            self.check_ecs(gb=param_points_grps, verbose=verbose)
            # turned this off temporarily...
            # FIX ME ("can only compare identically-labelled...")
            # (e.g. in 3.4.3 of nb8)
            
            # Generate self.probs if necessary
            if self.probs.is_empty:
                if verbose:
                    print('Initializing probability distribution...\n')
                # check that points are on a grid (the quick but slightly less certain way)
                param_lengths = [len(set(param_points[name])) for name in self.fit_param_names()]
                if not np.product(param_lengths)==len(param_points):
                    raise ValueError('Your modeled parameter space does not appear to be on a grid; the current version of bayesim can only handle initially gridded spaces (unless using a previously saved subdivided state).')
                    return
                else:
                    self.probs = Pmf(params=self.params.fit_params)

        elif mode=='function':
            # is there a way to save this (that could be saved to HDF5 too) so that subdivide can automatically call it?
            model_func = argv['model_data_func']

            # check that probs.points is populated...


            # iterate over parameter space and measured conditions to compute output at every point
            param_vecs = {p:[] for p in self.fit_param_names()}
            ec_vecs = {c:[] for c in self.ec_names()}
            model_vals = []

            for pt in self.probs.points.iterrows():
                param_vals = {p:pt[1][p] for p in self.fit_param_names()}
                for d in self.obs_data.iterrows():
                    ec_vals = {c:d[1][c] for c in self.ec_names()}
                    # add param and EC vals to the columns
                    for p in self.fit_param_names():
                        param_vecs[p].append(param_vals[p])
                    for c in self.ec_names():
                        ec_vecs[c].append(d[1][c])
                    # compute/look up the model data
                    # need to make sure that model_func takes in params and EC in appropriate format
                    model_vals.append(model_func(ec_vals,param_vals))

            # merge dictionaries together to put into a model data df
            vecs = deepcopy(param_vecs)
            vecs.update(ec_vecs)
            vecs.update({self.output_var:model_vals})

            self.model_data = pd.DataFrame.from_dict(vecs)

        # round EC's and generate groups
        rd_dct = {c.name:c.tol_digits for c in self.params.ecs}
        self.model_data = self.model_data.round(rd_dct)

        # round fit params - actually just force them to be members of the vals lists
        if verbose:
            print("Rounding model data...")
        for p in self.params.fit_params:
            #self.model_data[p.name] = [p.get_closest_val(val) for val in self.model_data[p.name]]
            if in_parallel:
                self.model_data[p.name] = Parallel(n_jobs=cpu_count())(delayed(get_closest_val)(val, p.vals) for val in self.model_data[p.name])
            else:
                md_temp = self.model_data[p.name]
                for i in range(len(md_temp)):
                    md_temp.iloc[i] = get_closest_val(md_temp.iloc[i], p.vals)
                self.model_data[p.name] = md_temp

        #print('At line 467...')
        #print(self.model_data.head())

        # drop any modeled data at EC's not measured
        # this was making mistakes previously so is commented out for now
        """
        if verbose:
            print("Removing modeled EC's that weren't measured...")
        self.model_data_ecgrps = self.model_data.groupby(self.ec_names())
        model_ecs = pd.DataFrame.from_records(data=list(self.model_data_ecgrps.groups.keys()), columns=self.ec_names()).sort_values(by=self.ec_names()).reset_index(drop=True)
        merged_ecs = model_ecs.merge(self.ec_pts, on=self.ec_names(), how='left', indicator=True)
        extra_pts = merged_ecs[merged_ecs._merge=='left_only']
        print('there are '+str(len(extra_pts))+' extra points...here are a few:')
        print(extra_pts.head())
        extra_data = merged_ecs[merged_ecs._merge=='right_only']
        if len(extra_data)>0:
            print("There seem to be data EC's that weren't modeled...")
            print(extra_data)
        print(extra_pts.index)
        self.model_data.drop(labels=extra_pts.index, axis=0, inplace=True)
        """
        #print('at line 485...')
        #print(self.model_data.head())
        
        # reset index to avoid duplication
        self.model_data.reset_index(inplace=True,drop=True)

        # (re-)generate groups
        if verbose:
            print("Grouping model data...")
        self.model_data_ecgrps = self.model_data.groupby(self.ec_names())
        self.model_data_grps = self.model_data.groupby(by=self.fit_param_names())

        # update flag and indices
        self.needs_new_model_data = False
        if verbose:
            print("Indexing model data...")
        self.calc_indices()

        if calc_model_unc:
            self.calc_model_unc(**argv)

    def comparison_plot(self,**argv):
        """
        Plot observed data vs. highest-probability modeled data.

        Args:
            ec_vals (`dict`): optional, dict of EC values at which to plot. If not provided, they will be chosen randomly. This can also be a list of dicts for multiple points.
            num_ecs (`int`): number of EC values to plot, defaults to 1 (ignored if ecs is provided)
            num_param_pts (`int`): number of the most probable parameter space points to plot (defaults to 1)
            ec_x_var (`str`): one of self.ec_names, will overwrite if this was provided before in attach_observations, required if it wasn't. If ec was provided, this will supercede that
            return_avg_err (bool): whether or not to return average (absolute) error for highest probability model over all EC's plotted (defaults to False)
            fpath (`str`): optional, path to save image to if desired
        """
        if 'ec_x_var' in argv.keys():
            self.params.set_ec_x(argv['ec_x_var'])

        if self.params.ec_x_name==None:
            print('You have not provided an x-variable from your experimental conditions against which to plot. Choosing the first one in the list, %s.\n'%(self.ec_names()[0]))
            self.params.set_ec_x(self.ec_names()[0])

        return_avg_err = argv.get('return_avg_err', False)
        all_errs = []

        other_ecs = [c for c in self.params.ecs if not c.name==self.params.ec_x_name]
        #print(self.params.ec_x_name, [c.name for c in other_ecs])

        # read in options and do some sanity checks
        num_ecs = argv.get('num_ecs',1)

        if len(other_ecs)==0: #only one experimental condition...
            one_ec = True
            ec_vals = [0] #placeholder, basically - this is klunky
        else: # more than one
            one_ec = False
            if 'ec_vals' in argv.keys():
                ec_vals = argv['ec_vals']
                if not (isinstance(ec_vals,list) or isinstance(ec_vals,np.ndarray)):
                    ec_vals = [ec_vals]
            else:
                ec_vals = []
                ec_pts = random.sample(list(self.model_data.groupby([c.name for c in other_ecs]).groups.keys()), num_ecs)

                for pt in ec_pts:
                    if len(other_ecs)==1: #pt will just be a float rather than a tuple
                        ec_vals.append({other_ecs[0].name:pt})
                    else:
                        ec_vals.append({other_ecs[i].name:pt[i] for i in range(len(pt))})

        num_param_pts = argv.get('num_param_pts',1)

        # need to fix this check
        #if self.params.ec_x_name in ecs.keys():
            #print('You provided a fixed value for your x variable, ignoring that and plotting the full range.')
            #del ecs[self.params.ec_x_name]

        # get parameter points for which to plot data
        param_pts = self.probs.most_probable(num_param_pts)

        # set up subplots
        fig, axs = plt.subplots(len(ec_vals), 2, figsize=(13,4*len(ec_vals)), squeeze=False)

        # get color cycle
        prop_cycle = plt.rcParams['axes.prop_cycle']
        colors = prop_cycle.by_key()['color']

        # at each set of conditions...
        for i in range(len(ec_vals)):
            obs_data = self.obs_data
            plot_title = 'Comparison  '
            if not one_ec:
                plot_title = plot_title[:-1] + 'at '
                ecs_here = ec_vals[i]
                # get obs data for correct EC values
                #print(ecs_here, obs_data.head())
                for c in other_ecs:
                    obs_data =  obs_data[abs(obs_data[c.name]-ecs_here[c.name])<=10.**(-1.*c.tol_digits)]
                    plot_title = plot_title + '%s=%s, '%(c.name,c.get_val_str(ecs_here[c.name]))
            obs_data = obs_data.sort_values(by=[self.params.ec_x_name]).reset_index(drop=True)

            # plot obs data
            axs[i,0].plot(obs_data[self.params.ec_x_name], obs_data[self.output_var], color=colors[0])

            legend_list = ['observed']
            c_ind = 1
            
            # get modeled data at these conditions, alternate way
            #md_ec_inds = [i for i in model_data.index if all([model_data.loc[i][list(d.keys())[0]]==list(d.values())[0] for d in ec_vals])]
            #model_data_thisec = deepcopy(self.model_data.loc[md_ec_inds])

            # plot modeled data and errors
            for pt in param_pts.iterrows():
                color = colors[c_ind]
                model_data = deepcopy(self.model_data.loc[self.model_data_grps.groups[tuple([pt[1][n] for n in self.fit_param_names()])]])
                if not one_ec:
                    # get modeled data just at these conditions
                    for c in other_ecs:
                        model_data =  model_data[abs(model_data[c.name]-ecs_here[c.name])<=10.**(-1.*c.tol_digits)]                           
                model_data.sort_values(by=[self.params.ec_x_name], inplace=True)
                model_data.reset_index(drop=True, inplace=True)
                # get the modeled values only at the observed points
                x_vals = list(obs_data[self.params.ec_x_name])
                model_data_drop = [i for i in range(len(model_data)) if model_data.loc[i][self.params.ec_x_name] not in x_vals]
                model_data.drop(labels=model_data_drop, axis=0, inplace=True)
                errors = np.subtract(model_data[self.output_var], obs_data[self.output_var])
                if c_ind==1: #take errors from highest probability
                    all_errs.extend([abs(e) for e in errors])

                #axs[i,0].plot(model_data[self.params.ec_x_name], model_data[self.output_var], color=color)
                #axs[i,1].plot(model_data[self.params.ec_x_name], errors, color=color)

                axs[i,0].plot(x_vals, model_data[self.output_var], color=color)
                axs[i,1].plot(x_vals, errors, color=color)
                
                leg_label = 'modeled: '
                for p in self.params.fit_params:
                    leg_label = leg_label + '%s=%s, '%(p.name, p.get_val_str(pt[1][p.name]))
                leg_label = leg_label[:-2]
                legend_list.append(leg_label)
                c_ind = c_ind+1

            # set ylims to match observed data and label axes
            obs_max = max(obs_data[self.output_var])
            obs_min = min(obs_data[self.output_var])
            obs_width = obs_max-obs_min
            axs[i,0].set_ylim([obs_min-0.05*obs_width, obs_max+0.05*obs_width])
            xvar = self.params.get_ec_x()
            axs[i,0].set_xlabel('%s [%s]' %(xvar.name, xvar.units))
            axs[i,1].set_xlabel('%s [%s]' %(xvar.name, xvar.units))
            yvar = self.params.find_param(self.output_var)
            axs[i,0].set_ylabel('%s [%s]' %(yvar.name, yvar.units))
            axs[i,1].set_ylabel('%s [%s]' %('$\Delta$'+yvar.name, yvar.units))
            axs[i,0].legend(legend_list, fontsize=14)
            plot_title = plot_title[:-2]
            axs[i,0].set_title(plot_title, fontsize=20)
            if one_ec:
                err_title = 'Errors'
            else:
                err_title = plot_title + ': errors'
            axs[i,1].set_title(err_title, fontsize=20)

            for j in [0,1]:
                for item in ([axs[i][j].xaxis.label, axs[i][j].yaxis.label] + axs[i][j].get_xticklabels() + axs[i][j].get_yticklabels()):
                    item.set_fontsize(18)

        plt.tight_layout()
        if 'fpath' in argv.keys():
            plt.savefig(argv['fpath'])

        if return_avg_err:
            return np.mean(all_errs)

    def run(self, **argv):
        """
        Do Bayes!
        Will stop iterating through observations if/when >= th_pm of probability mass is concentrated in <= th_pv of boxes and decide it's time to subdivide. (completely arbitrary thresholding for now)

        Args:
            save_step (`int`): interval (number of data points) at which to save intermediate PMF's (defaults to 10, 0 to save only final, <0 to save none)
            th_pm (`float`): threshold quantity of probability mass to be concentrated in th_pv fraction of parameter space to trigger the run to stop (defaults to 0.9)
            th_pv (`float`): threshold fraction of parameter space volume for th_pm fraction of probability to be concentrated into to trigger the run to stop (defaults to 0.05)
            min_num_pts (`int`): minimum number of observation points to use - if threshold is reached before this number of points has been used, it will start over and the final PMF will be the average of the number of runs needed to use sufficient points (defaults to 0.7 * the number of experimental measurements)
            prob_relax (`float`): number from 0 to 1.0, fraction of PMF from previous step to mix into prior for this step (defaults to 0) - higher values will likely converge faster but possibly have larger errors, especially if min_num_pts is small
            verbose (`bool`): flag for verbosity, defaults to False
        """
        # read in options
        save_step = argv.get('save_step', 10)
        th_pm = argv.get('th_pm', 0.9)
        th_pv = argv.get('th_pv', 0.05)
        min_num_pts = argv.get('min_num_pts', int(0.7*len(self.obs_data)))
        verbose = argv.get('verbose', False)
        rel = argv.get('prob_relax', 0.0)
        if rel < 0 or rel > 1.0:
            print("Relaxation parameter must be between 0 and 1.0 - defaulting to 0.")
            rel = 0
        if verbose:
            print('Running inference!\n')

        if min_num_pts > len(self.obs_data):
            print('Cannot use more observation points than there are in the data. Setting min_num_pts to len(self.obs_data)=%d\n'%len(self.obs_data))
            min_num_pts = len(self.obs_data)

        # do some sanity checks
        if self.needs_new_model_data:
            raise ValueError('Oops, you need to attach model data before running!')
            return

        if self.is_run:
            print('Running again at the same subdivision level. Previous results may be overridden...\n')

        # set up folders for intermediate files
        if save_step>0:
            folder = os.getcwd()
            pmf_folder = folder+'/PMFs/'
            obs_list_folder = folder+'/obs_lists/'
            for fp in [pmf_folder,obs_list_folder]:
                if not os.path.isdir(fp):
                    os.mkdir(fp)

        # rather than uniformizing, averaging the previous probs with a quasi-uniform
        old_probs = deepcopy(self.probs)
        uni_probs = deepcopy(self.probs)
        uni_probs.uniformize()
        start_probs = deepcopy(self.probs)
        start_probs.points['prob'] = rel*old_probs.points['prob'] + (1-rel)*uni_probs.points['prob']
        start_probs.normalize()

        # randomize observation order
        self.obs_data = self.obs_data.sample(frac=1)
        num_pts_used = 0
        num_runs = 0
        probs_lists = []
        delta_count_list = []
        nan_count_list = []
        skip_total = 0
        while num_pts_used < min_num_pts:
            prev_used_pts = num_pts_used
            num_runs = num_runs + 1
            obs_indices = []
            self.probs = deepcopy(start_probs)
            done=False
            
            while not done:
                # check if we've gone through all the points
                if num_pts_used == len(self.obs_data):
                    print('Used all the observed data! Last PMF to go into average may have been further from threshold condition.\n')
                    done = True
                else:
                    # get observed and modeled data
                    obs = self.obs_data.iloc[num_pts_used]
                    #print(obs)
                    obs_indices.append(num_pts_used)
                    ec = obs[self.ec_names()]
                    if len(self.ec_names())>1:
                        ecpt = tuple([ec[n] for n in self.ec_names()])
                    else:
                        ecpt = float(ec)
                    #print(ecpt)
                    #print(self.model_data_ecgrps.groups.keys())
                    model_here = deepcopy(self.model_data.loc[self.model_data_ecgrps.groups[ecpt]])
                    #print(len(model_here))

                    # compute likelihood and do a Bayesian update
                    #print(obs, model_here)
                    #print()
                    lkl, delta_count, nan_count, skip_count = self.probs.likelihood(meas=obs, model_at_ec=model_here, output_col=self.output_var, verbose=verbose)
                    self.probs.multiply(lkl)
                    num_pts_used = num_pts_used + 1
                    delta_count_list.append(delta_count)
                    nan_count_list.append(nan_count)
                    skip_total = skip_total + skip_count

                    # save intermediate PMF if necessary – the key is never used so it doesn't matter what it is
                    if save_step >0 and (num_pts_used-prev_used_pts) % save_step == 0:
                        self.probs.points.to_hdf(pmf_folder+'sub%d_run%d_PMF_%d.h5'%(self.probs.num_sub,num_runs,num_pts_used-prev_used_pts), "dummykey")

                    # check if threshold probability concentration has been reached
                    if np.sum(self.probs.most_probable(int(th_pv*len(self.probs.points)))['prob'])>th_pm:
                        done = True
                        print('done, points so far: '+str(num_pts_used))
                        
                if done:
                    probs_lists.append(np.array(self.probs.points['prob']))
                    if save_step >= 0:
                        self.probs.points.to_hdf(pmf_folder+'sub%d_run%d_PMF_final.h5'%(self.probs.num_sub,num_runs), "dummykey")
                    at_threshold=True
                    self.obs_data.iloc[obs_indices].to_hdf(obs_list_folder+'sub%d_run%d_obs_list.h5'%(self.probs.num_sub,num_runs), "dummykey")

    
        probs = np.mean(probs_lists,axis=0)
        self.probs.points['prob'] = probs
        print('Did a total of %d runs to use a total of %d observations.\n'%(num_runs,num_pts_used))

        print('\nAn average of %d / %d probability points had larger model uncertainty than experimental uncertainty during this run.\n'%(int(round(np.mean(delta_count_list))),len(self.probs.points)))

        print('\nAn average of %.2f / %d probability points were affected by missing/NaN simulation data.\n' %(np.mean(nan_count_list), len(self.probs.points)))
        print('\n%d points were skipped.\n' %skip_total)

        if save_step >=0:
            self.probs.points.to_hdf(pmf_folder+'sub%d_PMF_final.h5'%(self.probs.num_sub), "dummykey")
        self.is_run = True

    def subdivide(self, **argv):
        """
        Subdivide the probability distribution and save the list of new sims to run to a file.

        Args:
            threshold_prob (`float`): minimum probability of box to (keep and) subdivide (default value is the uniform distribution probability)
            new_sim_list_fpath (`str`): filename for file containing list of new simulations to be run (optional)
        """
        threshold_prob = argv.get('threshold_prob',1.0/len(self.probs.points))
        filename = argv.get('new_sim_list_fpath', 'new_sim_points_%d.h5'%(self.probs.num_sub))
        new_boxes = self.probs.subdivide(threshold_prob)
        #dropped_inds = list(dropped_boxes.index)

        #self.fit_params = [p for p in self.probs.params]
        self.attach_fit_params(self.probs.params)

        # remove old model data
        self.model_data = pd.DataFrame()

        # update flags
        self.needs_new_model_data = True
        self.is_run = False
        self.list_model_pts_to_run(fpath=filename)
        print('New model points to simulate are saved in the file %s.'%filename)

    def list_model_pts_to_run(self, fpath, **argv):
        """
        Generate full list of model points that need to be run (not just parameter points but also all experimental conditions). Saves to HDF5 at fpath.

        Note that this could be very slow if used on the initial grid (i.e. for potentially millions of points) - it's better for after a subdivide call.

        Args:
            fpath (`str`): path to save the list to (HDF5)
            verbose (`bool`): flag for verbosity, defaults to False
        """
        verbose = argv.get('verbose', False)
        if verbose:
            print('Listing the sets of simulation parameters that need to be run...\n')

        # get just the columns with the parameters
        param_pts = self.probs.points[self.fit_param_names()]

        # Now at every point in that list, make a row for every EC point
        param_inds = range(len(param_pts))
        ec_inds = range(len(self.ec_pts))
        columns = self.fit_param_names() + self.ec_names()
        pts = []
        for ppt in param_pts.iterrows():
            for ecpt in self.ec_pts.iterrows():
                pts.append(list(ppt[1])+list(ecpt[1]))
        sim_pts = pd.DataFrame(data=pts,columns=columns)
        sim_pts.to_hdf(fpath, "dummykey")

    def calc_model_unc(self, **argv):
        """
        Calculates largest difference in modeled output along any parameter direction for each experimental condition, to be used for uncertainty in calculating likelihoods. Currently only works if data is on a grid.

        (also assumes it's sorted by param names and then EC's)

        Args:
            verbose (`bool`): flag for verbosity, defaults to False
            model_unc_factor (`float`): multiplier on deltas to give uncertainty, defaults to 0.5 - smaller probably means faster convergence, but also higher chance to miss "hot spots"
            take_average (`bool`): flag for whether to use average model uncertainty at each measurement condition or the parameter-resolved version. Defaults to True if any parameters are logarithmically spaced, and False otherwise.
            min_unc_frac (`float`): minimum uncertainty as a fraction of the output variable value, defaults to 0.01
            min_unc_val (`float`): minimum uncertainty as an absolute number, defaults to 0.0
            in_parallel (`bool`): defaults to True, set False to force run in serial

        Note:
            If both `min_unc_frac` and `min_unc_val` are specified, the uncertainty will be set to the larger of the two in each case

        """

        factor = argv.get('model_unc_factor', 0.5)
        verbose = argv.get('verbose', False)
        if verbose:
            print('Calculating model uncertainty...')
        min_unc_frac = argv.get('min_unc_frac', 0.01)
        min_unc_val = argv.get('min_unc_val', 0.0)
        in_parallel = argv.get('in_parallel', True)
        if platform.system()=='Windows': #joblib doesn't work on Windows
            in_parallel = False
        
        # check if any parameters are log-spaced
        has_log = False
        if any([p.spacing=='log' for p in self.params.fit_params]):
            has_log = True
        take_average = argv.get('take_average', has_log)

        param_lengths = [p.length for p in self.params.fit_params]

        start_time = time.time()

        deltas = np.zeros(len(self.model_data))

        if not in_parallel:
            # need to do it in serial
            deltas_list = []
            for grp in self.model_data_ecgrps.groups:
                deltas_list.append(calc_deltas(grp, self.model_data_ecgrps.groups[grp], param_lengths, self.model_data, self.fit_param_names(), self.probs, self.output_var, take_average))
        else:
            # parallelize!
            deltas_list = Parallel(n_jobs=cpu_count())(delayed(calc_deltas)(grp, self.model_data_ecgrps.groups[grp], param_lengths, self.model_data, self.fit_param_names(), self.probs, self.output_var, take_average) for grp in self.model_data_ecgrps.groups)


        for entry in deltas_list:
            inds = self.model_data_ecgrps.groups[entry[0]]
            deltas[inds] = entry[1]

        # check for NaNs
        if np.any(np.isnan(deltas)):
            # there is a problem
            raise ValueError("Some uncertainties came out to NaN! Help!")
            
        # multiply by the factor
        self.model_data['uncertainty'] = factor * deltas

        # adjust upward for any minimum values
        unc = []
        for row in self.model_data.iterrows():
            min_unc = max(min_unc_val, min_unc_frac*row[1][self.output_var])
            unc.append(max(min_unc, row[1]['uncertainty']))
        self.model_data['uncertainty'] = unc

        if verbose:
            print('Calculating model uncertainty took %.2f seconds.\n'%(time.time()-start_time))

    def save_state(self,filename='bayesim_state.h5'): #rewrite this!
        """
        Save the entire state of this model object to an HDF5 file so that work can be resumed later.
        """

        # construct a dict with all the state variables
        state = {}

        # parameters
        state['params'] = self.params.as_dict()
        state['ec_pts'] = self.ec_pts
        state['output_var'] = self.output_var

        # PMF
        state['probs'] = self.probs.as_dict()

        # model/data
        state['model_data'] = self.model_data
        state['model_data_grps'] = self.model_data_grps
        state['model_data_ecgrps'] = self.model_data_ecgrps
        state['needs_new_model_data'] = self.needs_new_model_data
        state['obs_data'] = self.obs_data
        state['is_run'] = self.is_run

        # save the file
        state.to_hdf(filename, "dummykey")

    def visualize_grid(self,**argv):
        """
        Visualize the current state of the grid.

        Args:
            same as pmf.visualize()
        """
        self.probs.visualize(just_grid=True,**argv)

    def visualize_probs(self,**argv):
        """
        Visualize the PMF with a corner plot.

        Args:
            same as pmf.visualize()
        """
        self.probs.visualize(**argv)

# maybe I'll finish this eventually
    """
    def visualize_model_unc(self,**argv):

        #Visualize model uncertainty.

        #Args:
        #    ec_vals (`dict`): optional, dict of EC values at which to plot. If not provided, they will be chosen randomly. This can also be a list of dicts for multiple points.
        #    num_ecs (`int`): number of EC values to plot, defaults to 1 (ignored if ecs is provided)

        # read in options and do some sanity checks
        num_ecs = argv.get('num_ecs',1)
        if 'ec_vals' in argv.keys():
            ec_vals = argv['ec_vals']
            if not (isinstance(ec_vals,list) or isinstance(ec_vals,np.ndarray)):
                ec_vals = [ec_vals]
        else:
            ec_vals = []
            ec_pts_df = self.ec_pts.sample(num_ecs)
            ec_pts=[]
            for pt in ec_pts_df.iterrows():
                ec_vals.append({name:pt[1][name] for name in self.ec_names()})
                ec_pts.append(tuple(pt[1][self.ec_names()]))

        # set up subplots
        fig, axs = plt.subplots(len(ec_vals), figsize=(13,4*len(ec_vals)), squeeze=False)

        for i in range(len(ec_vals)):
            ecs_here = ec_vals[i]
            plot_title = ''
            for c in self.params.ecs:
                plot_title = plot_title + '%s=%s, '%(c.name,c.get_val_str(ecs_here[c.name]))
            subset = self.model_data.loc[self.model_data_ecgrps[tuple([ecs_here[n] for n in self.ec_names()])]]
            dense_grid = probs.populate_dense_grid(df=subset, col_to_pull='uncerainty', make_ind_lists=False)
            mat = dense_grid['mat']
            axs[i].imshow(mat)
            axs[i].set_title(plot_title)
    """

    def top_probs(self, num):
        """Return a DataFrame with the 'num' most probable points and some of the less interesting columns hidden."""
        df = self.probs.most_probable(num)
        cols = ['prob'] + [c for cs in [[p.name, p.name+'_min', p.name+'_max'] for p in self.params.fit_params] for c in cs]
        return df[cols]

    def set_param_info(self, param_name, **argv):
        """
        Set additional info for parameter param_name (any type).

        Args:
            param_name (str): name of parameter to modify
            units (str): units of parameter
            min_width (float): minimum width of parameter (only for fitting params)
            display_name (str): name to use on plots (can include TeX)
            tolerance (float): tolerance for this parameter
        """
        assert param_name in [p.name for p in self.params.all_params()], "I can't set info for a parameter (%s) that doesn't exist!"%param_name

        if 'units' in argv.keys():
            for i in range(len(self.params.fit_params)):
                if self.params.fit_params[i].name==param_name:
                    self.params.fit_params[i].units = argv['units']
            for i in range(len(self.params.ecs)):
                if self.params.ecs[i].name==param_name:
                    self.params.ecs[i].units = argv['units']
            for i in range(len(self.params.output)):
                if self.params.output[i].name==param_name:
                    self.params.output[i].units = argv['units']

        if 'min_width' in argv.keys():
            for i in range(len(self.params.fit_params)):
                if self.params.fit_params[i].name==param_name:
                    self.params.fit_params[i].min_width = argv['min_width']

        if 'display_name' in argv.keys():
            for i in range(len(self.params.fit_params)):
                if self.params.fit_params[i].name==param_name:
                    self.params.fit_params[i].display_name = argv['display_name']
            for i in range(len(self.params.ecs)):
                if self.params.ecs[i].name==param_name:
                    self.params.ecs[i].display_name = argv['display_name']
            for i in range(len(self.params.output)):
                if self.params.output[i].name==param_name:
                    self.params.output[i].display_name = argv['display_name']

        # pass these through to PMF since it doesn't always seem to happen automatically...
        if param_name in self.probs.param_names():
            for i in range(len(self.probs.params)):
                if self.probs.params[i].name==param_name:
                    if 'units' in argv.keys():
                        self.probs.params[i].units = argv['units']
                    if 'min_width' in argv.keys():
                        self.probs.params[i].min_width = argv['min_width']
                    if 'display_name' in argv.keys():
                        self.probs.params[i].display_name = argv['display_name']


    def ec_names(self):
        """Return list of experimental condition names."""
        return self.params.param_names('ec')

    def fit_param_names(self):
        """Return list of fitting parameter names."""
        return self.params.param_names('fit')



import numpy as np
import pandas as pd
import math
from scipy.stats import norm,lognorm
from itertools import product
from copy import deepcopy
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import bayesim.params as pm
import timeit
from itertools import product

class Pmf(object):
    """
    Class that stores a PMF capable of nested sampling / "adaptive mesh refinement".

    Stores probabilities in a DataFrame which associates regions of parameter space with probability values.
    """

    def make_points_list(self, params, total_prob=1.0):
        """
        Helper function for Pmf.__init__ as well as Pmf.subdivide.
        Given names and values for parameters, generate DataFrame listing
        values, bounds, and probabilities.

        Args:
            params (:class:`.Param_list`):
            total_prob (float): total probability to divide among points in parameter space - for an initialization, this is 1.0, for a subdivide call will be less.

        Returns:
            :obj:DataFrame with columns for each parameter's value, min, and max
            as well as a probability associated with that point in parameter
            space
        """

        # generate the list of box center coordinates
        param_indices = [range(len(p.vals)) for p in params]
        points = [p for p in product(*param_indices)]
        # iterate through to get values and add columns for param ranges
        for i in range(len(points)): # for every point (read: combination of value indices) in parameter space...
            row = []
            for j in range(len(params)): # for each parameter...
                # get value
                val_index = points[i][j]
                row.append(params[j].vals[val_index])
                # get min and max of bounding box
                row.extend([params[j].edges[val_index],params[j].edges[val_index+1]])
            points[i] = row

        points = np.array(points)
        columns = [c for l in [[n,n+'_min',n+'_max'] for n in self.param_names()] for c in l] # there has to be a more readable way to do this

        df = pd.DataFrame(data=points, columns=columns)
        df['prob'] = [total_prob/len(df) for i in range(len(df))]

        return df


    def __init__(self, **argv):
        """
        Initialize a Pmf object. Provide one of the three input options, or else will initialize empty.

        Args:
            params (:obj:`Param_list`): Param_list object containing parameters to be fit and associated metadata
            param_points (:obj:`DataFrame`): DataFrame containing all parameter points to start with
            prob_dict (:obj:`dict`): output of as_dict()
        """

        if 'prob_dict' in argv.keys():
            prob_dict = argv['prob_dict']
            self.is_empty = prob_dict['is_empty']
            self.num_sub = prob_dict['num_sub']
            self.points = prob_dict['points']
            self.params = [pm.Fit_param(**p) for p in prob_dict['params']]
        elif 'params' in argv.keys():
            # for now just copy in the param_list wholesale
            # eventually should probably scrub and/or update vals...
            self.num_sub = 0
            self.params = argv['params']
            self.points = self.make_points_list(self.params)
            if len(self.params)==0:
                self.is_empty = True
            else:
                self.is_empty = False
        elif 'param_points' in argv.keys():
            raise NotImplementedError('Still need to implement initializing a Pmf from a list of points!')
        else: # empty arguments
            self.num_sub = 0
            self.params = []
            self.points = pd.DataFrame()
            self.is_empty = True

    def as_dict(self):
        """Return this Pmf object in (readable) dictionary form."""
        d = deepcopy(self.__dict__)
        d['params'] = [p.__dict__ for p in self.params]
        return d

    def normalize(self, **argv):
        """Normalize overall PMF."""
        norm_const = self.points['prob'].sum()
        verbose = argv.get('verbose', False)
        #print(norm_const,type(norm_const))
        if abs(norm_const)<1e-12:
            if verbose:
                print("Somehow the normalization constant was zero! To play it safe, I won't do anything.")
        else:
            self.points['prob'] = [p/norm_const for p in self.points['prob']]

    def uniformize(self):
        """
        Keep PMF shape and subdivisions but make every probability equal.
        Useful for rerunning whole inference after subdividing.

        Note that because subdivisions are not uniform that this is NOT a uniform prior anymore.
        """
        norm_const = len(self.points)
        self.points['prob'] = [1.0/norm_const for i in range(norm_const)]

    def all_current_values(self, param):
        """
        List all values currently being considered for `param`.
        """
        ls = list(set(list(self.points[param])))
        ls.sort()
        return ls

    def find_neighbor_boxes(self, index):
        """
        Find and return all boxes neighboring the box at `index`.
        """
        all_vals = {name:self.all_current_values(name) for name in self.param_names()}
        param_spacing = {param.name:param.spacing for param in self.params}
        this_point = self.points.iloc[index]
        indices_to_intersect=[]
        # get range of values of each param to consider "neighbors"
        for p in self.params:
            param = p.name
            this_param_val = this_point[param]
            this_param_index = all_vals[param].index(this_param_val)

            # handle the edge cases
            if not this_param_index == 0:
                #down_delta = (all_vals[param['name']][this_param_index]-all_vals[param['name']][this_param_index-1])*1.001
                if param_spacing[param]=='linear':
                    down_delta = (this_point[param+'_max']-this_point[param+'_min'])*1.501
                elif param_spacing[param]=='log':
                    down_delta = this_point[param]-this_point[param+'_min']*(this_point[param+'_min']/(this_point[param+'_max']*1.001))
            else:
                down_delta=0
            if not this_param_index == len(all_vals[param])-1:
                #up_delta = (all_vals[param['name']][this_param_index+1]-all_vals[param['name']][this_param_index])*1.001
                if param_spacing[param]=='linear':
                    up_delta = (this_point[param+'_max']-this_point[param+'_min'])*1.501
                elif param_spacing[param]=='log':
                    up_delta = this_point[param+'_max']*(this_point[param+'_max']*1.001/this_point[param+'_min'])-this_point[param]
            else:
                up_delta=0
            #print(param,up_delta,down_delta)
            gt = self.points[param]>=this_param_val-down_delta
            lt = self.points[param]<=this_param_val+up_delta
            this_set = self.points[gt & lt]
            #print(this_set[[pm['name'] for pm in self.params]])
            indices_to_intersect.append(set(this_set.index.values))

        indices = list(set.intersection(*indices_to_intersect))
        indices.sort()
        neighbor_points = self.points.iloc[indices]
        # check if we went too far out along any axis
        inds_to_drop = []
        all_out_query = ''
        for p in self.params:
            param = p.name
            other_params = [pm.name for pm in self.params if not pm.name==param]
            # clunky brute-force search but it should be a smallish list so hopefully it won't kill us
            # first check that we're inside the box for all other params
            in_box_query = ''
            for q in other_params:
                if param_spacing[q]=='linear':
                    in_box_query = in_box_query + '%s<%f & %s>%f & '%(q,this_point[q+'_max'],q,this_point[q+'_min'])
                elif param_spacing[q]=='log':
                    in_box_query = in_box_query + '%s<%E & %s>%E & '%(q,this_point[q+'_max'],q,this_point[q+'_min'])
            # then check each direction in this param
            if param_spacing[param]=='linear':
                gt_query = in_box_query + '%s>%f'%(param,this_point[param+'_max'])
                lt_query = in_box_query + '%s<%f'%(param,this_point[param+'_min'])
                all_out_query = all_out_query + '(%s>%f | %s<%f) & '%(param,this_point[param+'_max'],param,this_point[param+'_min'])
            elif param_spacing[param]=='log':
                gt_query = in_box_query + '%s>%E'%(param,this_point[param+'_max'])
                lt_query = in_box_query + '%s<%E'%(param,this_point[param+'_min'])
                all_out_query = all_out_query + '(%s>%E | %s<%E) & '%(param,this_point[param+'_max'],param,this_point[param+'_min'])
            # pull points that are outside bounds of this param
            gt_check = neighbor_points.query(gt_query)
            lt_check = neighbor_points.query(lt_query)
            # group them by coords in other params
            gt_check_grps = gt_check.groupby(by=other_params)
            lt_check_grps = lt_check.groupby(by=other_params)
            # if any group has multiple indices, drop the ones corresponding to the "further out" values
            for grp,inds in gt_check_grps:
                if len(inds)>1:
                    vals = gt_check.loc[inds.index][param]
                    keep = vals.idxmin()
                    drop = [i for i in inds.index if not i==keep]
                    inds_to_drop.extend(drop)
                    #print('dropping '+str(drop))
            for grp,inds in lt_check_grps:
                if len(inds)>1:
                    vals = lt_check.loc[inds.index][param]
                    keep = vals.idxmax()
                    drop = [i for i in inds.index if not i==keep]
                    inds_to_drop.extend(drop)
                    #print('dropping '+str(drop))
        # and finally, check for ones that aren't inside bounds at all
        # for now, drop all of them - should keep closest corner really
        all_out_query = all_out_query[:-3]
        all_out_check = neighbor_points.query(all_out_query)
        inds_to_drop.extend(all_out_check.index)
        return neighbor_points.drop(labels=inds_to_drop)

    def subdivide(self, threshold_prob, include_neighbors=True, **argv):
        """
        Subdivide all boxes with P > threshold_prob and assign "locally uniform" probabilities within each box. If include_neighbors is true, also subdivide all boxes neighboring those boxes.

        Boxes with P < threshold_prob are deleted.

        Args:
            threshold_prob (`float`): probability above which a box should be retained
            include_neighbors (`bool`): whether to also subdivide all immediate neighbors to boxes meeting the threshold
            verbose (`bool`): Verbosity, default false
        """
        num_divs = {p.name:2 for p in self.params} #dummy for now

        verbose = argv.get('verbose', False)
        
        # pick out the boxes that will be subdivided
        to_subdivide = self.points[self.points['prob']>threshold_prob]
        #print(len(to_subdivide))
        #dropped_boxes = deepcopy(to_subdivide)

        num_high_prob_boxes = len(to_subdivide)

        if num_high_prob_boxes == 0:
            print('nothing to subdivide!')
            return

        if include_neighbors:
            # find neighbor boxes
            neighbor_list = []
            for box in to_subdivide.iterrows():
                neighbor_list.append(self.find_neighbor_boxes(box[0]))
            neighbors = pd.concat(neighbor_list)
            to_subdivide = pd.concat([to_subdivide,neighbors])
            to_subdivide = to_subdivide.drop_duplicates(subset=self.points.columns[self.points.columns != 'prob']) # exclude probability when considering identical-ness

        num_nbs = len(to_subdivide)-num_high_prob_boxes

        # check if minimum width is already satisfied for any parameter
        test_box = to_subdivide.iloc[0]
        for p in self.params:
            if p.spacing=='linear':
                this_width = box[1][p.name+'_max']-box[1][p.name+'_min']
            elif p.spacing=='log':
                this_width = box[1][p.name+'_max']/box[1][p.name+'_min']
            if this_width <= p.min_width:
                # don't divide along this direction
                num_divs[p.name] = 1
                if verbose:
                    print('Minimum width/factor of %s already satisfied for parameter %s, not subdividing in that direction!'%(p.get_val_str(p.min_width), p.name))

        # create new boxes (and delete old ones)
        new_boxes = []
        #dropped_inds = []
        for box in to_subdivide.iterrows():
            # first, remove this box from DataFrame
            #dropped_inds.append(box[0])
            self.points.drop([box[0]], inplace=True)

            # create new DataFrame with subdivided boxes
            new_pl = pm.Param_list()
            for p in self.params:
                # copy same params except for ranges
                new_pl.add_fit_param(name=p.name,
                val_range=[box[1][p.name+'_min'], box[1][p.name+'_max']], length=num_divs[p.name], min_width=p.min_width, spacing=p.spacing, units=p.units, tolerance=p.tolerance)
            # make new df, spreading total prob from original box among new smaller ones
            new_boxes.append(self.make_points_list(new_pl.fit_params, total_prob=box[1]['prob']))

        # put in the new points (and completely drop the old ones)
        self.points = pd.concat(new_boxes)

        # make new lists of self.params (this way might be slow...)
        new_params = pm.Param_list()
        for p in self.params:
            p_args = p.__dict__
            p_args['vals'] = vals=list(set(list(self.points[p.name])))
            del p_args['val_range']
            #print(p_args)
            #new_params.add_fit_param(name=p.name, vals=list(set(list(self.points[p.name]))), spacing=p.spacing)
            new_params.add_fit_param(**p_args)
        self.params = [p for p in new_params.fit_params]

        # sort values
        self.points = self.points.sort_values(self.param_names())

        # reindex DataFrame
        self.points = self.points.reset_index(drop=True)

        # should be normalized already, but just in case:
        self.normalize(verbose=verbose)

        # increment subdivide count
        self.num_sub = self.num_sub + 1

        print(str(num_high_prob_boxes) + ' box(es) with probability > ' + str(threshold_prob) + ' and ' + str(num_nbs) + ' neighboring boxes subdivided!')

        return new_boxes

    def multiply(self, other_pmf, **argv):
        """
        Compute and store renormalized product of this Pmf with other_pmf.

        Args:
            other_pmf (:class:`.Pmf`): PMF to multiply by
        """

        verbose = argv.get('verbose', False)
        
        # check for silliness
        assert isinstance(other_pmf, Pmf), "You didn't feed in a Pmf object!"
        assert len(self.points) == len(other_pmf.points), "Pmf's are over different numbers of points. Can't exactly do a pointwise multiplication on that, can I?"
        # should add a check that all points match

        # copy and sort/index DataFrames to match so that we're multiplying
        # the right probabilities together
        these_probs = deepcopy(self.points)
        other_probs = deepcopy(other_pmf.points)

        these_probs.sort_values(by=self.param_names())
        other_probs.sort_values(by=self.param_names())

        # actually multiply
        new_probs = these_probs['prob'] * other_probs['prob']

        # check if there's any probability there...
        if abs(np.sum(new_probs))<1e-12:
            if verbose:
                print("You're gonna have a bad time...not multiplying")
        else:
            self.points['prob'] = new_probs
            self.normalize(verbose=verbose)

    def likelihood(self, **argv):
        """
        Compute likelihood over this Pmf's parameter space given modeled data at the given EC's for every parameter space point and a measurement at the same EC's.

        Args:
            meas (`float`): one output value
            model_at_ec (:py:class:`DataFrame`): DataFrame containing model data at the experimental condition of the measurement and uncertainty values in a column called 'uncertainty' for every point in parameter space
            output_col (`str`): name of column with output variable

            ec: dict with keys of condition names and values
            meas: one output value e.g. J
            unc: uncertainty in measured value (stdev of a Gaussian)
            model_func: should accept one dict of params and one of conditions and output measurement (might deprecate)
            verbose (`bool`): be verbose or not
        """

        # read in and process inputs
        meas = argv['meas']
        model_data = argv['model_at_ec']
        model_data.sort_values(self.param_names())
        model_data.reset_index(drop=True)
        output_col = argv['output_col']
        meas_val = meas[output_col]
        meas_err = meas['uncertainty']
        verbose = argv.get('verbose', False)

        # set up likelihood DF
        lkl = deepcopy(self)
        new_probs = np.zeros([len(lkl.points),1])

        delta_count = 0
        nan_count = 0
        skip_count = 0
        
        # check that lengths match
        if len(lkl.points)==len(model_data):
            # here's the actual loop that computes the likelihoods
            for point in lkl.points.iterrows():
                #print(ec, point[1])
                #model_val = model_func(ec, dict(point[1]))
                model_pt = model_data.iloc[point[0]]
                model_val = float(model_pt[output_col])
                if not np.isnan(model_val):
                    model_err = float(model_pt['uncertainty'])
                    if np.isnan(model_err):
                        err = 2*meas_err #janky but whatev
                    else:
                        err = model_err + meas_err

                    # tally how many times deltas were bigger
                    if model_err > meas_err:
                        delta_count = delta_count + 1

                    new_probs[point[0]] = norm.pdf(meas_val, loc=model_val, scale=abs(err))
                else:
                    new_probs[point[0]] = 1.0/len(lkl.points)
                    nan_count = nan_count + 1
            
            # copy these values in
            lkl.points['prob'] = new_probs

            # make sure that the likelihood isn't zero everywhere...
            if abs(np.sum(new_probs))<1e-12:
                if verbose:
                    print('likelihood has no probability! :(')
                skip_count = skip_count + 1
            if any(np.isnan(np.array(self.points['prob']))):
                print(self.points[self.points.isnull().any(axis=1)])
                raise ValueError('Uh-oh, some probability is NaN!')

        else:
            if verbose:
                print('Missing model data for ')
                print(meas)
                print('Skipping likelihood calculation here.\n')

        lkl.normalize()        
        return lkl, delta_count, nan_count, skip_count

    def most_probable(self, n):
        """Return the n largest probabilities in a new DataFrame.
        """
        sorted_probs = self.points.sort_values(by='prob',ascending=False)
        return sorted_probs.iloc[0:n]

    def param_names(self):
        """Return list of parameter names of this PMF."""
        return [p.name for p in self.params]

    def populate_dense_grid(self,**argv):
        """
        Populate a grid such as the one created by make_dense_grid.

        Args:
            df (`obj`:DataFrame): DataFrame to populate from (should have columns for every param)
            col_to_pull (`str`): name of the column to use when populating grid points
            make_ind_lists (`bool`): whether to return a list of indices corresponding to the first in every slice (used by bayesim.model.calc_model_gradients)

        Returns:
            a dict with keys for each thing requested
        """
        # read in inputs
        df = argv['df']
        col_to_pull = argv['col_to_pull']
        make_ind_lists = argv.get('make_ind_lists',True)

        # initialize things
        mat_shape = []
        pvals_indices = {}
        param_edges = {}

        for param in self.params:
            mat_shape.append(param.length)
            pvals_indices[param.name] = {param.vals[i]:i for i in range(len(param.vals))}
            #indices_lists.append(range(len(param_vals[pname])))

        mat = np.full(mat_shape,np.nan)

        # initialize optional things
        if make_ind_lists:
            ind_lists = {p.name:[] for p in self.params}

        for pt in df.iterrows():
            #print(pt)
            slices = []
            param_point = self.points.loc[pt[0]] # if df isn't self.points
            for p in self.params:
                min_val = param_point[p.name+'_min']
                max_val = param_point[p.name+'_max']
                inds = [pvals_indices[p.name][v] for v in p.vals if v>min_val and v<max_val]
                #print(p.name, min_val, max_val, p.vals, inds)
                slices.append(slice(min(inds),max(inds)+1,None))
                if make_ind_lists:
                    ind_lists[p.name].append(inds[0])
            if col_to_pull == 'prob':
                val = param_point['prob']
            else:
                val = pt[1][col_to_pull]
            mat[tuple(slices)] = val

        return_dict = {'mat':mat}
        if make_ind_lists:
            return_dict['ind_lists'] = ind_lists

        return return_dict

    def weighted_avgs(self):
        """Returns a dict with the weighted average for each parameter."""
        avgs = {}
        for param in self.params:
            bins, probs = self.project_1D(param)
            vals = param.vals
            avgs[param.name] = np.sum([probs[i]*vals[i] for i in range(len(vals))])
        return avgs


    def project_1D(self, param, dense_grid=[]):

        """
        Project down to a one-dimensional PMF over the given parameter. Used by the visualize() method.

        Args:
            param (:obj:`Fit_param`): one of `self.params`

        Returns:
            bins (:obj:`list` of :obj:`float`): bin edges for plotting with matplotlib.pyplot.hist (has length one more than next return list)
            probs (:obj:`list` of :obj:`float`): probability values for histogram-style plot - note that these technically have units of the inverse of whatever the parameter being plotted is (that is, they're probability densities)
            dense_grid (matrix): optionally, pass precomputed dense grid to save time
        """
        ## first find bin edges
        # pull all bounds, then flatten, remove duplicates, and sort
        bins = sorted(list(set(list(self.points[param.name+'_min'])+list(self.points[param.name+'_max']))))

        if dense_grid==[]:
            # generate dense grid and populate with probabilities
            dense_grid = self.populate_dense_grid(df=self.points, col_to_pull='prob', make_ind_lists=False)
            mat = dense_grid['mat']
        else:
            mat = dense_grid

        # sum along all dimensions except the parameter of interest
        param_ind = self.param_names().index(param.name)
        inds_to_sum_along = tuple([i for i in range(len(mat.shape)) if not i==param_ind])
        probs = np.nansum(mat,axis=inds_to_sum_along)

        return bins, probs

    def project_2D(self, x_param, y_param, no_probs=False, dense_grid=[]):

        """
        Project down to two dimensions over the two parameters. This one doesn't actually need to sum, it just draws a bunch of (potentially overlapping) rectangles with transparencies according to their probability densities (as a fraction of the normalized area). Used by the visualize() method.

        Args:
            x_param (dict): one of `self.params`, to be the x-axis of the 2D joint plot
            y_param (dict): one of `self.params`, to be the y-axis of the 2D joint plot
            no_probs (bool): whether to just show the grid boxes or include probability as determiner of transparency
            dense_grid (matrix): optionally, pass precomputed dense grid to save time

        Returns:
            (:obj:`list` of :obj:`matplotlib.patches.Rectangle`): patches for plotting the 2D joint probability distribution
        """
        if dense_grid==[]:
            # generate dense grid and populate with probabilities
            dense_grid = self.populate_dense_grid(df=self.points, col_to_pull='prob', make_ind_lists=False)
            mat = dense_grid['mat']
        else:
            mat = dense_grid

        max_prob = max(self.points['prob'])
        #param_vals = dense_grid['param_vals']
        param_edges = {p.name:p.edges for p in self.params}

        # sum along all dimensions except the parameter of interest
        inds_to_sum_along = tuple([i for i in range(len(mat.shape)) if not self.params[i].name in [x_param.name,y_param.name]])
        dense_probs = np.nansum(mat, axis=inds_to_sum_along)

        if no_probs:
            # generate list of patch parameters - first need every pair of indices
            ind_pairs = product(*[range(i) for i in dense_probs.shape])
            patches = []
            for pr in ind_pairs:
                x_min = param_edges[x_param.name][pr[0]]
                x_width = param_edges[x_param.name][pr[0]+1] - x_min
                y_min = param_edges[y_param.name][pr[1]]
                y_width = param_edges[y_param.name][pr[1]+1] - y_min
                if not dense_probs[pr]==np.nan:
                    patches.append(mpl.patches.Rectangle((x_min,y_min), x_width, y_width, fill=False, ec='k'))

            return patches

        else:
            return dense_probs

    def visualize(self, **argv):
        """
        Make histogram matrix to visualize the PMF.

        Args:
            frac_points (`float`): number >0 and <=1 indicating fraction of total points to visualize (will take the most probable, defaults to 1.0)
            just_grid (`bool`): whether to show only the grid (i.e. visualize subdivisions) or the whole PMF (defaults to False)
            fpath (`str`): optional, path to save image to
            true_vals (`dict`): optional, set of param values to highlight on PMF
            return_plots (bool): whether to return figure and axes (used by visualize_PMF_sequence in utils), default False
            color_index (int): from 0 to 4, choose from blue, green, red, purple, orange (defaults to 0)
        """
        # read in options
        frac_points = argv.get('frac_points', 1.0)
        just_grid = argv.get('just_grid', False)
        return_plots = argv.get('return_plots', False)
        if return_plots:
            plt.ioff()
            data = {i:{} for i in range(len(self.params))}
        color_index = argv.get('color_index',0)

        if 'fpath' in argv.keys():
            fpath = argv['fpath']

        if 'true_vals' in argv.keys():
            # check that all params are there
            true_vals = argv['true_vals']
            if not set(true_vals.keys())==set(self.param_names()):
                print('Your true_vals do not have all the paramter names! Proceeding without them.')
                plot_true_vals = False
            # check that values are within ranges
            elif not all(true_vals[p.name]>p.edges[0] and true_vals[p.name]<p.edges[-1] for p in self.params):
                print('Your true_vals are not within the correct bounds. Proceeding without them.')
                plot_true_vals = False
            else:
                plot_true_vals = True
        else:
            plot_true_vals = False

        # get color cycle and colormaps (just pull five)
        prop_cycle = plt.rcParams['axes.prop_cycle']
        colors = prop_cycle.by_key()['color'][:5]
        cmap_name_list = ['Blues','Greens','Reds','Purples','Oranges']
        color = colors[color_index]
        cmap_name = cmap_name_list[color_index]

        start_time = timeit.default_timer()

        # find ranges to plot - this likely needs tweaking
        N = len(self.points)
        points_to_include = self.most_probable(int(frac_points*N))
        plot_ranges = {}
        for param in self.params:
            plot_ranges[param.name] = [min(points_to_include[param.name+'_min']), max(points_to_include[param.name+'_max'])]

        fig, axes = plt.subplots(nrows=len(self.params), ncols=len(self.params), figsize=(5*len(self.params),5*len(self.params)))

        check1 = timeit.default_timer()
        time1 = round(check1-start_time,2)
        #print('setup finished in ' + str(time1) + ' seconds')

        dense_probs = self.populate_dense_grid(df=self.points, col_to_pull='prob', make_ind_lists=False, return_edges=True)['mat']

        param_ticks = self.pick_ticks(plot_ranges)

        for rownum in range(0,len(self.params)):
            for colnum in range(0,len(self.params)):
                x_param = self.params[colnum]
                y_param = self.params[rownum]

                # pre-formatting
                x_min = plot_ranges[x_param.name][0]
                x_max = plot_ranges[x_param.name][1]
                axes[rownum][colnum].set_xlim(x_min, x_max)
                axes[rownum][colnum].set_axisbelow(True)

                for item in ([axes[rownum][colnum].xaxis.label, axes[rownum][colnum].yaxis.label] +axes[rownum][colnum].get_xticklabels() + axes[rownum][colnum].get_yticklabels()):
                    item.set_fontsize(20)

                if rownum==colnum: #diagonal - single-variable histogram
                    if just_grid:
                        fig.delaxes(axes[rownum][colnum])
                    else:
                        diag_start = timeit.default_timer()
                        bins, probs = self.project_1D(x_param, dense_grid=deepcopy(dense_probs))
                        checkpoint = round(timeit.default_timer()-diag_start,2)
                        #print('project_1D took ' + str(checkpoint) + ' seconds')
                        if x_param.spacing=='log':
                            axes[rownum][colnum].set_xscale('log')
                            xtick_locs = [10**tick for tick in param_ticks[x_param.name]['locs']]
                            vals = [math.sqrt(bins[i]*bins[i+1]) for i in range(len(probs))]
                        elif x_param.spacing=='linear':
                            xtick_locs = param_ticks[x_param.name]['locs']
                            vals = [0.5*(bins[i]+bins[i+1]) for i in range(len(probs))]

                        # plot the data
                        axes[rownum][colnum].hist(vals, weights=probs, bins=bins, edgecolor='k', linewidth=1.0, color=color)
                        if return_plots:
                            data[rownum][colnum]={'vals':vals,'weights':probs,'bins':bins,'color':color}

                        # formatting...
                        # turn the ticks off and on (was having some stray minor ticks show up sometimes)
                        axes[rownum][colnum].tick_params(axis='x',which='both',bottom=False,labelbottom=False)
                        axes[rownum][colnum].tick_params(axis='x',which='major',bottom=True,labelbottom=True)
                        axes[rownum][colnum].set_xticks(xtick_locs, minor=False)
                        axes[rownum][colnum].set_xticklabels(param_ticks[x_param.name]['labels'])
                        axes[rownum][colnum].set_ylim(0,1)
                        axes[rownum][colnum].yaxis.set_label_position("right")
                        axes[rownum][colnum].yaxis.tick_right()
                        for label in axes[rownum][colnum].get_yticklabels():
                            label.set_fontsize(20)
                        axes[rownum][colnum].set_ylabel('P(%s)'%x_param.display_name, rotation=270, labelpad=20) #labelpad is kind of a brute-force way to do this and might break if we change the figure size, but va='bottom' wasn't working
                        #axes[rownum][colnum].`grid`(axis='y')

                        # add true value if desired
                        if plot_true_vals:
                            true_x = [true_vals[x_param.name]]
                            axes[rownum][colnum].scatter(true_x,0.05,200,'#FFFF00',marker='*',zorder=100)

                        diag_finish = timeit.default_timer()
                        diag_time = round(diag_finish-diag_start,2)
                        #print('diagonal plot finished in ' + str(diag_time) + ' seconds')

                elif rownum > colnum: # below diagonal
                    offdiag_start = timeit.default_timer()
                    # formatting axes ranges and ticks
                    if x_param.spacing=='log':
                        axes[rownum][colnum].set_xscale('linear')
                        x_min = np.log10(plot_ranges[x_param.name][0])
                        x_max = np.log10(plot_ranges[x_param.name][1])
                        axes[rownum][colnum].set_xlim([x_min, x_max])
                    if y_param.spacing=='log':
                        y_min = np.log10(plot_ranges[y_param.name][0])
                        y_max = np.log10(plot_ranges[y_param.name][1])
                    else:
                        y_min = plot_ranges[y_param.name][0]
                        y_max = plot_ranges[y_param.name][1]
                    axes[rownum][colnum].set_xticks(param_ticks[x_param.name]['locs'])
                    axes[rownum][colnum].set_xticklabels(param_ticks[x_param.name]['labels'])
                    axes[rownum][colnum].set_yticks(param_ticks[y_param.name]['locs'])
                    axes[rownum][colnum].set_yticklabels(param_ticks[y_param.name]['labels'])
                    axes[rownum][colnum].set_ylim([y_min, y_max])


                    if just_grid:
                        patches = self.project_2D(x_param, y_param, no_probs=True)
                        axes[rownum][colnum].grid(False)
                        for patch in patches:
                            axes[rownum][colnum].add_patch(patch)
                    else:
                        dense_probs_here = self.project_2D(x_param, y_param, dense_grid=deepcopy(dense_probs))


                        # generating the colormap and transparency
                        cmap = plt.get_cmap(cmap_name)
                        probs = dense_probs_here.transpose()
                        scaled_probs = probs/np.amax(probs)
                        alpha_weight = 0.5+0.1*color_index
                        alphas = alpha_weight*scaled_probs + (1-alpha_weight)*probs
                        colors = mpl.colors.Normalize(0.0, 1.0)(alphas)
                        colors = cmap(colors)
                        colors[..., -1] = alphas
                        # plot the data
                        axes[rownum][colnum].imshow(colors, aspect='auto', origin='lower', extent=(x_min, x_max, y_min, y_max), zorder=1)
                        if return_plots:
                            data[rownum][colnum]={'image':colors,'extent':(x_min, x_max, y_min, y_max)}

                    checkpoint = round(timeit.default_timer()-offdiag_start,2)
                    #print('project_2D took ' + str(checkpoint) + ' seconds')

                    if plot_true_vals:
                        true_x = [true_vals[x_param.name]]
                        true_y = [true_vals[y_param.name]]
                        if x_param.spacing=='log':
                            true_x = np.log10(true_x)
                        if y_param.spacing=='log':
                            true_y = np.log10(true_y)
                        axes[rownum][colnum].scatter(true_x,true_y,200,c="None",marker='o',linewidths=3,edgecolors='#FFFF00',zorder=100)
                    #axes[rownum][colnum].grid(False)
                    offdiag_finish = timeit.default_timer()
                    offdiag_time = round(offdiag_finish-offdiag_start,2)
                    #print('off-diagonal plot finished in ' + str(offdiag_time) + ' seconds')

                else: # above diagonal
                    fig.delaxes(axes[rownum][colnum])

        # put the labels on the outside
        for i in range(0,len(self.params)):
            xlabel = '%s [%s]' %(self.params[i].display_name, self.params[i].units)
            ylabel = '%s [%s]' %(self.params[i].display_name, self.params[i].units)
            axes[len(self.params)-1][i].set_xlabel(xlabel)
            if i>0: # top one is actually a probability
                axes[i][0].set_ylabel(ylabel)

        plt.tight_layout()

        if 'fpath' in argv.keys():
            plt.savefig(fpath)

        if return_plots:
            return {'fig':fig, 'axes':axes, 'data':data}
            plt.close()

        else:
            plt.show()


    def pick_ticks(self, plot_ranges):
        """Helper function for visualize."""
        param_ticks = {}
        for p in self.params:
            param_ticks[p.name] = {'locs':[], 'labels':[]}
            min_val = plot_ranges[p.name][0]
            max_val = plot_ranges[p.name][1]
            if p.spacing=='linear':
                p_range = max_val-min_val
                tick_spacing = round(p_range/4.5, -1*int(round(np.log10(0.1*p_range))))
                first_tick = int(min_val/tick_spacing)*tick_spacing
                if first_tick < min_val:
                    first_tick = first_tick+tick_spacing
                tick_locs = np.arange(first_tick, 0.01*p_range+max_val, tick_spacing)
                param_ticks[p.name]['locs'] = tick_locs
                if tick_spacing>2:
                    param_ticks[p.name]['labels'] = [str(int(loc)) for loc in tick_locs]
                else:
                    num_digits = int(round(-np.log10(tick_locs[-1]-tick_locs[0])+1))
                    param_ticks[p.name]['labels'] = [str(round(loc,num_digits)) for loc in tick_locs]
            elif p.spacing=='log':
                p_range = np.log10(max_val)-np.log10(min_val)
                if p_range<1.0:
                    tick_spacing=0.2
                elif p_range<3.0:
                    tick_spacing = 0.5
                elif p_range<=6.0:
                    tick_spacing = 1.0
                else:
                    tick_spacing = 2.0
                first_tick = int(np.log10(min_val)/tick_spacing)*tick_spacing
                if first_tick < np.log10(min_val):
                    first_tick = first_tick + tick_spacing
                tick_locs = np.arange(first_tick, 0.01*p_range+np.log10(max_val), tick_spacing)
                param_ticks[p.name]['locs'] = tick_locs
                if tick_spacing<1.0:
                    param_ticks[p.name]['labels'] = [r'10$^{%s}$'%str(round(loc,1)) for loc in tick_locs]
                else:
                    param_ticks[p.name]['labels'] = [r'10$^{%s}$'%int(loc) for loc in tick_locs]
        return param_ticks



from setuptools import setup, find_packages
import os, sys

setup(name='bayesim',
      version='0.10.0',
      description='Fast model fitting via Bayesian inference',
      author='Rachel Kurchin, Giuseppe Romano',
      author_email='rkurchin@mit.edu, romanog@mit.edu',
      classifiers=['Programming Language :: Python :: 3.7'],
      long_description=open('README.rst').read(),
      install_requires=['pandas', 'scipy', 'matplotlib', 'joblib', 'pytables'],
      license='GPLv2',
      packages = ['bayesim'],
      entry_points = {
     'console_scripts': [
      'bayesim=bayesim.__main__:main'],
      },
      zip_safe=False)
